"""

Implementation of the wave field solution of 1D and 2D waves. For the Wave1D and Wave2D
class, a power spectral density distribution function :math:`S(k)` can be imposed on
the wave nodes such that we can simulate the propagation of waves obeying either a
Jonswap or Gauss spectrum.

The theoretical background has the following contents

* Wave1D_ : Description of the Wave1D class
    - dft1d_ :  Discrete Fourier Transform implementation for 1D wave spectra
    - fft1d_ :  Fast Fourier Transform implementation for 1D waves spectra
* Wave2D_ : Description of the Wave2D class
    - dft2d_ :  Discrete Fourier Transform implementation for 2D wave spectra
    - fft2d_ :  Fast Fourier Transform implementation for 2D wave spectra
* References_ : List of literature used for this implementation


.. _Wave1D:

=============================================
Wave1D: Theoretical background on one-D waves
=============================================

.. _dft1d:

Wave equation with DFT
----------------------

Assume a spectrum :math:`S(k_i)`, with :math:`i=0, n_k`. The wave field
:math:`\\eta(\\mathbf{x}, t)` at time :math:`t` along spatial positions :math:`x_p`
(where :math:`p=0, n_x`) is constructed using a Discrete Fourier Transform (DFT) as
follows

.. math ::

    \\eta(x_p, t) = \\sum_i^{n_k} a_i \\exp(j (k_i x_p - \\omega_i t + \\phi_i))

where :math:`j^2\\equiv -1`, the amplitude :math:`a_i` follows from the power spectral
density :math:`S(\\mathbf{k})` as

.. math ::

    a_i = \\sqrt{2 S(k_i) dk_i},

:math:`\\phi_i` is the random phase corresponding to the wave node :math:`i`, and the
angular frequency :math:`\\omega_i` is related to the wave vector via the deep water
dispersion relation according to

.. math ::

    \\omega_i = \\sqrt{g  |k_i|}

with :math:`g=9.81 m/s^2` being the gravity constant.
We can rewrite the wave equation in matrix form:

.. math ::

    \\eta(x_p, t) = \\sum_i^{n_k} M_{ip}\\exp(-j \\omega_i t)

where the matrix :math:`M_{ip}` is

.. math ::

    M_{ip} = \\hat{A}_i \\exp(j k_i x_p)

and the complex  amplitude :math:`\\hat{A}_i` follows from

.. math ::

    \\hat{A}_{i} = a_i\\exp(j\\phi_i)

From the matrix form is can be seen that the wave equation can be calculated by a dot
product multiplication of the :math:`n_x \\times n_k` matrix :math:`M_{ip}` and the
vector :math:`\\exp(-j\\omega_i t)` of size :math:`n_k \\times 1`.
The result is a :math:`1 \\times n_x` vector of the wave along the
x-axis :math:`\\eta(x_p, t)`. This is how the DFT is implemented in the code of Wave1D.

Note that we have described the wave equation in terms of the power spectral density
(PSD) equation in the wave vector domain :math:`k`, denoted here as :math:`S(k)`.
Normally the PSD is given in the angular frequency domain :math:`\\omega` as
:math:`S^\\prime(\\omega)`
We can obtain the version in wave vector domain using the equality

.. math ::

    S(k) = S^\\prime(\\omega)\\frac{d\\omega}{dk} =
           S^\\prime(\\sqrt{g k}) \\frac{g}{2 k}.


Using the PSD in the wave vector domain has the advantage that we only have to deal with
the periodicity of the discrete Fourier transform in the spatial domain x: a wave will
be period after a length :math:`2\\pi/\\Delta k`. To get rid of this periodicity is we
can simply use a spatial domain which larger than this repetition length.

.. _fft1d:

Wave equation with FFT
----------------------

In case the number of wave numbers nodes is equal to the number of spatial nodes, and we
have ensured that

.. math ::

    \\hat{A}(-k_i)\\exp(-j\\omega t) = \\hat{A}^\\ast(k_i)\\exp(j\\omega t)

we can use the FFT algorithm for calculating the wave equation at time :math:`t`. This
is because the FFT uses the symmetry rule that for any *real* valued signal
:math:`F(x)`, the Fourier transform :math:`\\hat{F}(k)`  by definition has the property
that :math:`\\hat{F}(-k) = \\hat{F}^\\ast(k)` (where the :math:`\\ast` indicates the
complex conjugate). The FFT is a  much more efficient way to obtain a solution for the
wave equation: the calculation time for a DFT  of a signal with :math:`N` nodes is
proportional to :math:`N^2`, while the calculation time of an FFT is proportional to
:math:`N\\log(N)`. The FFT is used for the Wave1D solution when the *wave_construction*
field is set to *FFT*.

.. _Wave2D:

=============================================
Wave2D: Theoretical background on two-D waves
=============================================

The Wave2D Class gives a description of a 2D propagating wave with a spectral density
distribution given by either a Gauss or a Jonswap spectrum. The class extents the
functionality of the Wave1D: it obtains all the information related to the 1D wave in
k-vector domain from the wave1D attribute field which stores a Wave1D object. For all
the information additionally required for the 2D description (like spreading function),
extra attributes are added to the class.

Again, three ways to solve the wave field from the spectral components have been
implemented:

1. *DFTpolar*: A 2D wave field is constructed from the multiplication of the Spectral
   density S(k) and the directional distribution D(theta). A Discrete Fourier Transform
   (DFT) is used to calculate the wave field. This is a straightforward implementation,
   but again, DFT is `slow` to calculate, especially in 2D, so only use this for
   testing. You can speed up the calculation by selecting wave node with an amplitude
   larger than a certain threshold.
2. *FFT*: if symmetry in spectral k domain is imposed, we can again use the FFT to
   calculate the wave field. Recommended
3. *DFTcartesian*: The exact same symmetric spectral amplitudes as the FFT is used, but
   the wave is calculated with a DFT. This is so slow that it is not possible to
   include all wave components, so only used for testing purposes.

.. _dft2d:

2D DFT implementation
---------------------

A 2D wave can be described by its spectral density distribution :math:`S(k)` and the
directional distribution :math:`D(\\theta)` as

.. math ::

    E_{k, \\theta}(k, \\theta) = S(k)D(\\theta)

where for :math:`S(k)` we can either take a Jonswap or a Gauss distribution as
implemented in the *wave_spectra* module in *spectrum_jonswap* and *spectrum_gauss*.
Since we are working in the wave vector domain :math:`k`, we must first transform the
spectrum from :math:`\\omega` to :math:`k` space, just as described in the 1D wave
theory. For directional distribution, the *spreading_function* is recommended.

The *DFTpolar* implementation constructs the complex amplitudes by multiplying the
spectral density :math:`E_{k, \\theta}` with its bin width :math:`dk d\\theta`:

.. math ::

    \\hat{A}_{k, \\theta} =
        \\sqrt{2 E_{k, \\theta}(k, \\theta) dk d\\theta} \\exp({j\\phi})

where the phase :math:`\\phi` is a random number between :math:`0\\sim 2\\pi`.
The wave amplitude is then calculate by simply summing over all the wave components we
have added, similar to our 1D version

.. math ::

    \\eta(\\mathbf{x}, t) =
    \\sum_p^{n_{k}}\\sum_q^{n_{\\theta}} \\hat{A}_{k_p, \\theta_q}
    \\exp(j (\\mathbf{k}_{p,q}\\cdot\\mathbf{x}- \\omega(k_p) t))

where :math:`\\mathbf{k} = (k_x, k_y)` is the 2D wave vector in cartesian coordinates
belonging to the sample wave vector in polar coordinates :math:`(k_p, \\theta_q)` with
:math:`k=|\\mathbf{k}|`. The wave vector magnitude, :math:`\\mathbf{x}` is the cartesian
vector  in spatial domain.  Also we are using the deep water dispersion relation again
to calculate the angular wave frequency :math:`\\omega` for a given wave
vector :math:`k`.

The DFT implementation can be sped up significantly by smartly selecting nodes, for
instance based on the components above a certain threshold, or we could make a
non-uniform distribution. Both methods have been implemented and can be selected by
settings the *wave_selection* attribute field to  *Subrange* or *EqualEnergyBins*,
respectively.

.. _fft2d:

2D FFT implementation
---------------------

As already mentioned, the DFT implementation is very slow. We can speed up by selecting
wave nodes in a smart way, but this consequently also takes out information from our
resulting wave field. On top of that: it turns out that even after selecting only the
most important spectral wave components, the DFT still is not very fast.

The recommended way to construct a wave field is again by using the FFT setting for the
*wave_construction* argument. There is only one catch: the FFT algorithm requires the
spectral components described in the cartesian wave vector domain, whereas so far we
have been working in the polar wave vector domain. On top of that, we have to impose
the symmetry required for a FFT:

.. math ::

    \\hat{A}(-\\mathbf{k})\\exp(-j\\omega t)=
        \\hat{A}^\\ast(\\mathbf{k})\\exp(j\\omega t)

This symmetry rule is the 2D version from the one we have seen in the 1D wave section.

For the FFT method we need a description of the complex amplitudes in the
Cartesian domain, i.e. we need to derive :math:`E_{k}(\\mathbf{k})` from our starting
point `E_{k, \\theta}(k, \\theta)`. It can be shown that  we need to do

.. math ::

    E_{k}(\\mathbf{k}) = \\frac{E_{k, \\theta}(k, \\theta)}{k}

This equation is used in the function *spectrum2d_complex_amplitudes* as we divide the
directional spreading function with the wave vector magnitude. For a more detailed
derivation, please have a look at the References_.

.. _References:

References
----------

* Det Norske Veritas: Environmental conditions and environmental loads, April 2007,
  DNV-RP-C205
* Jocely Fr\'echot: Realistic simulation of ocean surface using wave spectra.
  Proceedings of the First International Conference on Computer Graphics Theory and
  Applications (GRAPP, 2006), 2006, Portugal, p 76--83 <hal-00307938>

=========================================
Start of the *wave_fields* implementation
=========================================

"""

import logging
from os.path import splitext

import colorcet as cc
import h5py
import matplotlib.animation as animation
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
import seaborn as sns
from matplotlib.colors import LightSource
from numpy.fft import fftshift
from scipy.constants import g as g0  # gravity constant 9.81 m/s2

import pymarine.waves.wave_spectra as ms
from pymarine.utils.coordinate_transformations import polar_to_cartesian
from pymarine.utils.numerical import find_idx_nearest_val
from pymarine.utils.plotting import clean_up_artists, set_limits

sns.set(context="notebook")

logger = logging.getLogger(__name__)


class PlotProperties:
    """
    A class containing the properties of the line plot.

    Parameters
    ----------
    show: bool, optional,
        Show the current wae
    color: str, optional
        Store the color of the line plotting the wave. Default = "w"
    linewidth: int, optional
        Width of the line plotting the wave. Default = 1
    scattersize: int, optional
        Size of the scatter points plotted at the wave. Default = 0, i.e., no scatter
        points are plotted

    Notes
    -----
    * Only used by WaveSimulator to connect plot properties to each wave via the *plot*
      attribute field of the wave
    * Not critical and perhaps this class should be moved out of the *wave_fields*
      module later. For now, leave it as it is. It doesn't hurt me as well.
    """

    def __init__(self, show=True, color="w", linewidth=1, scattersize=0):
        self.show = show
        self.color = color
        self.linewidth = linewidth
        self.scattersize = scattersize
        self.save_image = False
        self.save_data = False


def _energy_deficit(k, k0, E, Hs, Tp, gamma, spectral_version, spectrum_type, sigma):
    """
    Helper function for fsolve to calculate the integral of the spectrum between k0 and
    k1 and subtract E

    Parameters
    ----------
    k: ndarray
       Array with the wave vector  nodes
    k0: float
       Start integrating at wave vector k0
    E:
        Energy per bin
    Hs:
        Significant wave height
    Tp:
        Peak period
    gamma:
        Peakness of Jonswap
    spectral_version:
        Version of implementation used
    spectrum_type: {"jonswap", "gauss"}
        Type of wave spectrum. Default = jonswap
    sigma:
        Width of the Gauss

    Returns
    -------
    float:
        Energy deficit

    Notes
    -----

    This function is used a helper function for fsolve and should not be called
    directly. It is used to calculate the integrated energy of a bin between k and k+dk

    """

    # Solve the integral of the spectrum between k0 and k1 and subtract E and returns
    # the deficit energy.
    # This function can be used to calculate the end border of an integration interval
    # such that the area of the integral results in a given value E
    (delta_e, err) = (
        sp.integrate.quad(
            ms.spectrum_wave_k_domain,
            k0,
            k,
            args=(Hs, Tp, gamma, sigma, spectrum_type, spectral_version),
        )
        - E
    )
    return delta_e


class Wave2D:
    """
    A class for linearized solutions of the 2D wave (deep water, linear). The Wave1D is
    used to describe the radial direction

    Parameters
    ----------
    wave1D: :obj:`Wave1D`
        The Wave1D field to describe the spectral properties along the radial direction
        in k-domain the 2D wave
    Lx: float, optional
        Length of the domain in x-direction in m, default = 200 m
    Ly: float, optional
        Length of the domain in y-direction in m, default = 200 m
    xmin: float, optional
        Starting coordinate of the x-domain. Default = 0 m
    ymin: float, optional
        Starting coordinate of the y-domain. Default = 0 m
    nx_points: int, optional
        Number of nodes in the x-domain. Default = 64
    ny_points: int, optional
        Number of nodes in the y-domain. Default = 64
    kx_min: float, optional
        Start of wave vector domain. Default = 0 rad/m
    kx_max: float, optional
        End of wave vector domain. Default = pi rad/m
    delta_kx: float, optional
        Bin spacing in wave vector domain. Default is pi/32
    n_kx_nodes: int, optional
        Number of bins in in x direction in k vector domain. Default = 64
    ky_min: float, optional
        Start of wave vector domain. Default = 0 rad/m
    ky_max: float, optional
        End of wave vector domain. Default = pi rad/m
    delta_ky: float, optional
        Bin spacing in wave vector domain. Default is pi/32
    n_ky_nodes: int, optional
        Number of bins in y direction in k vector domain. Default = 64
    theta_min: float, optional
         Minimum value of theta direction in polar domain. Default = 0 rad
    theta_max: float, optional
         Maximum value of theta direction in polar domain. Default = 2 pi rad
    n_theta_nodes: int, optional
         Number of nodes in theta direction. Default = 100
    Theta_0: float, optional
         Main wave direction in rad. Default = 0 rad. The valid range is: -pi/2 < T < pi
    Theta_s_spreading_factor: float, optional
        Spreading factor (s-definition) in theta domain. Default = 5 (Typical for wind
        waves, for swell use 13.0)
    """

    def __init__(
        self,
        wave1D,
        name=None,
        Lx=200,
        Ly=200,
        xmin=0,
        ymin=0,
        nx_points=64,
        ny_points=64,
        kx_min=0.0,
        kx_max=3.1415,
        delta_kx=0.0981,
        n_kx_nodes=32,
        ky_min=0.0,
        ky_max=3.1415,
        delta_ky=0.0981,
        n_ky_nodes=32,
        theta_min=0,
        theta_max=2 * np.pi,
        n_theta_nodes=100,
        Theta_0=0,
        Theta_s_spreading_factor=5,
    ):
        logger.info("Initialise JonSwap 1D wave field")

        if name is None:
            self.name = "_".join(
                [
                    "wave",
                    wave1D.spectrum_type,
                    wave1D.wave_construction,
                    wave1D.wave_selection,
                    "2d",
                ]
            )
        else:
            self.name = name

        self.wave1D = wave1D
        self.Lx = Lx
        self.Ly = Ly
        self.nx_points = nx_points
        self.ny_points = ny_points

        self.xmin = xmin
        self.Lx = Lx
        self.xmax = self.xmin + self.Lx
        self.xmid = self.xmin + self.Lx / 2.0
        self.xpoints = np.linspace(self.xmin, self.xmax, self.nx_points)
        self.delta_x = self.xpoints[1] - self.xpoints[0]

        self.ymin = ymin
        self.Ly = Ly
        self.ymax = self.ymin + self.Ly
        self.ymid = self.ymin + self.Ly / 2.0
        self.ypoints = np.linspace(self.ymin, self.ymax, self.ny_points)
        self.delta_y = self.ypoints[1] - self.ypoints[0]

        self.n_theta_nodes = n_theta_nodes
        self.Theta_0 = Theta_0
        self.Theta_s_spreading_factor = Theta_s_spreading_factor

        self.theta_points = None
        self.kx_nyquist = None
        self.kt = None
        self.delta_theta = None

        # Here, theta=0 corresponds to the north direction. Rotate clock wise
        self.theta_min = theta_min
        self.theta_max = theta_max

        self.kx_min = kx_min
        self.kx_max = kx_max
        self.delta_kx = delta_kx
        self.n_kx_nodes = n_kx_nodes
        self.k_polar_mesh = None
        self.k_polar_bin_area_over_kk = None
        self.omega_dispersion = None
        self.E_wave_density_polar = None

        self.ky_min = ky_min
        self.ky_max = ky_max
        self.delta_ky = delta_ky
        self.n_ky_nodes = n_ky_nodes
        self.k_xy_mesh = None

        self.kx_nodes = None
        self.ky_nodes = None

        self.k_cartesian_mesh = None
        self.kk = None
        self.E_wave_complex_amplitudes = None
        self.omega_sign = None

        self.xy_mesh = None

        self.amplitude = None

        self.phase = None

        self.theta_area_fraction = 1

        # use seed to update the random phase if required
        self.seed = 1
        self.update_phase = True

        self.update_x_k_theta_sample_space()
        self.D_spread = np.zeros(self.theta_points.shape)
        self.calculate_spreading_function()
        self.calculate_spectral_components()
        self.calculate_wave_surface()

    def make_report(self):
        """Make report of settings for this wave"""

        frm = "{:40s} : {}"

        print(frm.format("Name", self.name))
        print("----------- Domain settings --------")
        print(frm.format("Start x [m]", self.xmin))
        print(frm.format("Start y [m]", self.ymin))
        print(frm.format("End x [m]", self.xmax))
        print(frm.format("End y [m]", self.ymax))
        print(frm.format("Length x [m]", self.Lx))
        print(frm.format("Length y [m]", self.Ly))
        print("----------- Time settings --------")
        print(frm.format("Start t [s]", self.wave1D.t_start))
        print(frm.format("End t [s]", self.wave1D.t_end))
        print(frm.format("Length t [s]", self.wave1D.t_length))
        print(frm.format("Delta t [s]", self.wave1D.delta_t))

        n_x_points_total = self.xpoints.size * self.ypoints.size
        n_k_points_total = self.k_cartesian_mesh[0].size

        print("----------- Numerical resolutions --------")
        print(frm.format("Number x - nodes", self.xpoints.size))
        print(frm.format("Number y - nodes", self.ypoints.size))
        print(frm.format("Total number of spatial points", n_x_points_total))
        if self.kx_nodes is not None:
            print("# Cartesian mesh specifications")
            print(frm.format("Number kx - nodes", self.kx_nodes.size))
            print(frm.format("Number ky - nodes", self.ky_nodes.size))
            print(frm.format("Delta x [m]", self.delta_x))
            print(frm.format("Delta y [m]", self.delta_y))
            dkx = np.diff(self.kx_nodes)
            dky = np.diff(self.ky_nodes)
            print(frm.format("Delta kx min [rad/m]", dkx.min()))
            print(frm.format("Delta ky min [rad/m]", dky.min()))
            print(frm.format("Delta kx max [rad/m]", dkx.max()))
            print(frm.format("Delta ky max [rad/m]", dky.max()))
            print(frm.format("Delta kx first [rad/m]", self.delta_kx))
            print(frm.format("Delta ky first [rad/m]", self.delta_ky))
            print(frm.format("kx-min [rad/m]", self.kx_nodes.min()))
            print(frm.format("ky-min [rad/m]", self.ky_nodes.min()))
            print(frm.format("kx-max [rad/m]", self.kx_nodes.max()))
            print(frm.format("ky-max [rad/m]", self.ky_nodes.max()))
            print(frm.format("kx-nyq [rad/m]", self.kx_nyquist))
            print(frm.format("ky-nyq [rad/m]", self.ky_nyquist))
            print(frm.format("Lx_max [m]", 2 * np.pi / dkx[0]))
            print(frm.format("Ly_max [m]", 2 * np.pi / dky[0]))
        else:
            print("# Polar mesh specifications")
            print(frm.format("Number k_r - nodes", self.k_polar_mesh[1].shape[0]))
            print(frm.format("Number theta - nodes", self.k_polar_mesh[0].shape[1]))
            print(frm.format("Number total nodes", n_k_points_total))

        print("----------- Numerical methods --------")
        print(frm.format("Selection method", self.wave1D.wave_selection))
        print(frm.format("Construction method", self.wave1D.wave_construction))
        print(frm.format("DFT N x N", n_x_points_total * n_k_points_total))
        print(frm.format("FFT N x log(N)", n_x_points_total * np.log(n_x_points_total)))

    def update_k_polar_mesh(self):
        """Update the polar mesh and its bin area divided by k belonging to the current
        k/theta mesh

        Notes
        -----
        The bin area follows from the theta and k linear array as k x dtheta x dk
        However,  to calculate the complex wave amplitude we need to divide D(omega, kk)
        by kk to go the cartesian mesh.
        Here, by not multiplying with kk to get the bin.
        Therefore, we calculate darea / kk = kk dtheta * dkk / kk = dtheta x dkk

        """
        # create polar mesh if direction x kk_magnitude and store in k_polar_mesh.
        # k_polar_mesh[0] are the directions over the 2D mesh, k_polar_mesh[1] are the
        # wave vector magnitudes over the 2D mesh
        self.k_polar_mesh = np.meshgrid(self.theta_points, self.wave1D.kx_nodes)

        # k_polar_mesh[0] is the theta over the 2D plane, so the delta theta is the
        # gradient-y
        delta_theta = np.gradient(self.k_polar_mesh[0])[1]
        # k_polar_mesh[1] is the kr over the 2D plane, so the delta kr is the gradient-x
        delta_k_r = np.gradient(self.k_polar_mesh[1])[0]

        # the  area of a polar mesh is k * dtheta * dk and dived by k, so the
        # kmagnitude drops here
        self.k_polar_bin_area_over_kk = delta_theta * delta_k_r

        # turn the 2D polar coordinates into cartesian coordinates convert the polar
        # angle from mathematical definition (theta=0 -> x-axis and counter-clock
        # rotation) to naval definition (theta=0 is y-axis and clock-wise rotation)
        # theta_naval=pi/2-theta
        self.k_cartesian_mesh = np.array(
            polar_to_cartesian(self.k_polar_mesh[1], np.pi / 2 - self.k_polar_mesh[0])
        )
        self.kk = np.sqrt(self.k_cartesian_mesh[0] ** 2 + self.k_cartesian_mesh[1] ** 2)

    def update_x_k_theta_sample_space(self):
        self.theta_points = np.linspace(
            self.theta_min, self.theta_max, self.n_theta_nodes
        )

        self.delta_theta = self.theta_points[1] - self.theta_points[0]

        self.xmax = self.xmin + self.Lx

        self.xmid = self.xmin + self.Lx / 2.0

        self.xpoints = np.linspace(self.xmin, self.xmax, self.nx_points)

        self.delta_x = self.xpoints[1] - self.xpoints[0]

        self.ymax = self.ymin + self.Ly

        self.ymid = self.ymin + self.Ly / 2.0

        self.ypoints = np.linspace(self.ymin, self.ymax, self.ny_points)

        self.delta_y = self.ypoints[1] - self.ypoints[0]

        self.xy_mesh = np.meshgrid(self.xpoints, self.ypoints, indexing="ij")

        self.amplitude = np.zeros(self.xy_mesh[0].shape)

        self.kx_nyquist = np.pi / self.delta_x
        self.ky_nyquist = np.pi / self.delta_y

        if self.wave1D.wave_construction == "DFTpolar":
            # take the (non)-uniform wave vectors from the 1D wave
            self.update_k_polar_mesh()

        else:
            # For the FFT, the number wave vectors should be equal to the number of x
            # points. For the DFT on the cartesian mesh, we use the same mesh as the
            # FFT, so we can compare the speed of the algorithms
            self.kx_nodes = 2 * np.pi * np.fft.fftfreq(self.nx_points, self.delta_x)
            self.ky_nodes = 2 * np.pi * np.fft.fftfreq(self.ny_points, self.delta_y)

            self.delta_kx = self.kx_nodes[1] - self.kx_nodes[0]
            self.delta_ky = self.ky_nodes[1] - self.ky_nodes[0]

            # create the mesh [KX, KY]
            self.k_xy_mesh = np.meshgrid(self.kx_nodes, self.ky_nodes, indexing="ij")
            self.k_cartesian_mesh = self.k_xy_mesh
            self.kk = np.sqrt(
                self.k_cartesian_mesh[0] ** 2 + self.k_cartesian_mesh[1] ** 2
            )

    def calculate_spreading_function(self):
        """Calculate the spreading function"""
        self.D_spread = ms.spreading_function(
            theta=self.theta_points,
            theta0=self.Theta_0,
            s_spreading_factor=self.Theta_s_spreading_factor,
        )

        (
            self.itheta_low,
            self.itheta_high,
            self.theta_low,
            self.theta_high,
            self.theta_peak,
            self.D_peak,
            self.D_spread_area,
            self.theta_mask,
        ) = ms.thetaspreadspecs(
            self.theta_points, self.D_spread, self.theta_area_fraction
        )

        # check if we need to mask the points
        mask_points = False
        if self.wave1D.wave_selection == "Subrange":
            mask_points = True
        if (
            self.wave1D.wave_selection == "EqualEnergyBins"
            and self.wave1D.use_subrange_energy_limits
        ):
            mask_points = True

        if mask_points:
            # extra wave vectors based on the range k_low,k_high
            self.theta_points = np.extract(self.theta_mask, self.theta_points)
            self.D_spread = np.extract(self.theta_mask, self.D_spread)

            if self.wave1D.sample_every > 1:
                self.theta_points = self.theta_points[:: self.wave1D.sample_every]
                self.D_spread = self.D_spread[:: self.wave1D.sample_every]

            # we need to update the k_polar_mesh as well
            self.update_k_polar_mesh()

        if self.phase is not None:
            if (
                self.phase.shape[0] != self.wave1D.n_kx_nodes
                or self.phase.shape[1] != self.theta_points.size
            ):
                phase_shape_mismatch = True
            else:
                phase_shape_mismatch = False
        if self.phase is None or phase_shape_mismatch or self.update_phase:
            self.E_wave_density_polar = np.zeros(self.k_cartesian_mesh[0].shape)
            self.phase = ms.initialize_phase(
                self.E_wave_density_polar, self.wave1D.seed
            )
            self.update_phase = False

    def calculate_spectral_components(self):
        """
        Calculate the 2D wave density function from the current k-array and spreading
        angles
        """

        if self.wave1D.wave_construction == "DFTpolar":
            # the DFT based on a polar mesh is used. Calculate the polar coordinates
            # and turn it in cartesian values

            self.omega_dispersion = np.sqrt(
                self.wave1D.gravity0 * abs(self.k_polar_mesh[1])
            ) * np.where(self.k_polar_mesh[1] >= 0, 1, -1)

            # The wave density function follows from Sk * Dspread.
            # To get a matrix out of it, multiply the transposed S^T with the D
            n_k = self.wave1D.spectrumK.size
            n_d = self.D_spread.size
            self.E_wave_density_polar = self.wave1D.spectrumK.reshape(
                n_k, 1
            ) * self.D_spread.reshape(1, n_d)

            # In the polar domain, the integral multiplied with delta_theta*delta_kx
            # give
            # the complex amplitude a_k = sqrt(2 S(k, theta) dk * dtheta
            self.E_wave_complex_amplitudes = np.sqrt(
                2 * self.E_wave_density_polar * self.k_polar_bin_area_over_kk
            ) * np.exp(1j * self.phase)
        else:
            # Either a DFT (wave_construction==DFTcartesian) or a FFT
            # (wave_construction==FFT) based on a cartesian mesh is used.
            self.k_cartesian_mesh = self.k_xy_mesh
            self.kk = np.sqrt(self.k_xy_mesh[0] ** 2 + self.k_xy_mesh[1] ** 2)
            self.kk = np.where(
                abs(self.kk) < ms.TINY, ms.TINY * np.ones(self.kk.shape), self.kk
            )

            (
                self.E_wave_complex_amplitudes,
                self.omega_sign,
            ) = ms.spectrum2d_complex_amplitudes(
                kx_nodes=self.kx_nodes,
                ky_nodes=self.ky_nodes,
                Hs=self.wave1D.Hs,
                Tp=self.wave1D.Tp,
                gamma=self.wave1D.gamma,
                Theta_0=self.Theta_0,
                Theta_s_spread_kx=self.Theta_s_spreading_factor,
                spectrum_type=self.wave1D.spectrum_type,
                spectral_version=self.wave1D.spectral_version,
            )

            k_bin_area = self.delta_kx * self.delta_ky
            # area_polar = self.wave1D.kx_nodes * self.k_polar_bin_area_over_kk
            np_e_sq = np.square(abs(self.E_wave_complex_amplitudes))
            self.E_wave_density_polar = np_e_sq / (k_bin_area / 2) / self.kk

            # calculate the omega values belong to the wave vectors
            self.calculate_omega_dispersion()

    def calculate_omega_dispersion(self):
        # Calculate the omega frequency belonging to the wave vectors according to the
        # deep water dispersion relation.
        self.omega_dispersion = np.sqrt(g0 * abs(self.kk)) * self.omega_sign
        self.delta_omega = np.diff(self.omega_dispersion)

    def calculate_wave_surface(self):
        if not self.wave1D.wave_construction == "FFT":
            # For the DFT directly calculate the wave field from the spectral components
            self.amplitude = self.dft_complex_amplitudes(
                self.E_wave_complex_amplitudes, self.omega_dispersion, self.wave1D.time
            )
            if self.wave1D.wave_construction == "DFTcartesian":
                # Scale the amplitude with a factor 2 because we used the two-side
                # k-space
                self.amplitude *= 0.5
        else:
            # get the wave field using an FFT
            self.amplitude = self.fft_amplitude(
                self.E_wave_complex_amplitudes, self.omega_dispersion, self.wave1D.time
            )

        logger.debug(f"H_s of 2D surface {4 * np.std(self.amplitude)}  ")

    def dft_complex_amplitudes(self, S_tilde, omega, time):
        """Calculate DFT of complex amplitudes at time 'time'

        Parameters
        ----------
        S_tilde: Complex amplitude obtained from
        omega: angular frequency for each node
        time: current time

        Returns
        -------
        ndarray:
            2D array with Discrete fourier transform

        Notes
        -----
        * The trick with the exponential matrix exp (j*x*k) does not work in 2D because
          you run out of memory too fast.
          Therefore, calculate the wave field with a loop over the wave vectors.
        * This algorithm is really slow, so you should use FFT for 2D waves!

        """

        dft = np.full(self.xy_mesh[0].shape, 0 + 0 * 1j, dtype=complex)
        KX = self.k_cartesian_mesh[0]
        KY = self.k_cartesian_mesh[1]
        for j in range(self.k_cartesian_mesh[0].shape[1]):
            for i in range(self.k_cartesian_mesh[0].shape[0]):
                s_theta = S_tilde[i, j]
                dft += s_theta * np.exp(
                    1j
                    * (
                        self.xy_mesh[0] * KX[i, j]
                        + self.xy_mesh[1] * KY[i, j]
                        - omega[i, j] * time
                    )
                )
        return np.real(dft)

    def fft_amplitude(self, S_tilde, omega, time):
        """
        Calculate Fourier transform of S using the FFT
        Parameters
        ----------
        S_tilde: ndarray, complex
            Complex array of the fouriern components
        omega: ndarray
            Angular freqyencies belong to the wave vectors
        time: float
            Current time

        Returns
        -------
        ndarray
            real array with the DFT of the complex amplitudes

        """
        N = int(S_tilde.size / 2)
        ampl = N * np.fft.ifft2(S_tilde * np.exp(1j * (-time * omega)))
        # ampl should be real already because S_tilde should be symmetrical around
        # k=0 S(k)=S^*(-k) to be sure, take the real value only
        return np.real(ampl)

    def export_complex_amplitudes(self, filename, exportAsHD5=True):
        """Export the calculated complex amplitudes to HDF 5 file

        Parameters
        ----------
        filename: str
            Name of the file
        exportAsHD5: bool, optional
            Export as HD5. Default = True. If false, the complex amplitudes are written
            to Ascii

        """
        if exportAsHD5:
            filebase, ext = splitext(filename)
            logger.debug(f"writing hd5 file to {filebase}")
            with h5py.File(filebase + ".h5", "w") as hf:
                hf.create_dataset("Kx", data=self.k_cartesian_mesh[0])
                hf.create_dataset("Ky", data=self.k_cartesian_mesh[1])
                hf.create_dataset("Amodulus", data=abs(self.E_wave_complex_amplitudes))
                hf.create_dataset(
                    "Aphase", data=np.angle(self.E_wave_complex_amplitudes)
                )
                hf.create_dataset("omega", data=self.omega_dispersion)

        else:
            # write the complex amplitudes to the file filename in ascii format
            logger.debug(f"writing numpy ascii file to {filename}")

            nx = self.E_wave_complex_amplitudes.shape[0]
            ny = self.E_wave_complex_amplitudes.shape[1]
            KX = self.k_cartesian_mesh[0]
            KY = self.k_cartesian_mesh[1]
            with open(filename, "w") as f:
                f.write(
                    "# complex amplitudes a at kx,ky mesh {}x{} kxtheta\n".format(
                        nx, ny
                    )
                )
                f.write(
                    "# {:>18s}{:>20s}{:>20s}{:>20s}{:>20s}\n".format(
                        "kx", "ky", "real(a)", "imag(a)", "omega"
                    )
                )
                for j in range(ny):
                    f.write(f"# j={j}\n")
                    for i in range(nx):
                        ampl = self.E_wave_complex_amplitudes[i][j]
                        f.write(
                            "{:20.8g}{:20.8g}{:20.8g}{:20.8g}{:20.8g}\n".format(
                                KX[i, j],
                                KY[i, j],
                                float(np.real(ampl)),
                                float(np.imag(ampl)),
                                float(self.omega_dispersion[i, j]),
                            )
                        )
                    f.write("\n")

    def propagate_wave(self):
        """Increase the time and recalculate the surface"""

        # the time is stored in the wave1D object
        self.wave1D.next_time()
        self.calculate_wave_surface()

    def plot_wave(
        self,
        figsize=None,
        r_axis_type="frequency",
        plot_title=None,
        x_plot_title=0.05,
        y_plot_title=0.99,
        min_data_value=None,
        max_data_value=None,
        number_of_contour_levels=10,
        title_font_size=10,
        x_time_label=0.05,
        y_time_label=0.95,
        x_hs_label=0.05,
        y_hs_label=0.92,
        add_hs_estimate=True,
        color_map=cc.m_coolwarm,
        zorder=0,
        use_contourf=True,
    ):
        """Create a plot of the current wave

        Parameters
        ----------
        figsize: tuple
            x and y size of the total figure. Default is None, so figure size is not
            imposed
        r_axis_type: {"frequency", "period"}, optional
            quantity to use along the radial axis. Either frequency [Hz] or period [s].
            Default = "frequency"
        plot_title: str
            Title to put in th figure. If None, use the titles found in the liftdyn file
        x_plot_title: float, optional
            x position (as a fraction of the sub plot index_title_axis). Default = 0.0
        y_plot_title: float, optional
            y position (as a fraction of the sub plot index_title_axis). Default = 1.1
        delta_y_titles: float, optional
            spacing between the lines of thet title lines. Default = 0.05
        max_title_lines: int or None
            Maxinmum number of title lines to plot. Default = None, ie. plot all
        title_font_size: int
            Size of the title font
        r_axis_lim: tuple or None
            The limits of the x axis. Default = None, so no limits are imposed
        zorder:  int, optional
            Position of the contour plot. Default = 0, meaning that the contours are
            placed at the bottom and the grid and labels are visible
        use_contourf: bool, optional
            If true, use contourf to make the contour plot. Slower. Default = False

        Returns
        -------
        tuple (fig, axis)
            Handle to the figure and the axis
        """
        if figsize is not None:
            size = figsize
        else:
            size = None

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=size)
        ax.set_aspect("equal", "datalim")

        x_label = "X [m]"
        y_label = "Y [m]"
        data_x_2d = self.xy_mesh[0]
        data_y_2d = self.xy_mesh[1]
        amplitude = self.amplitude

        if min_data_value is None:
            v_min = np.nanmin(amplitude)
        else:
            v_min = min_data_value

        if max_data_value is None:
            v_max = np.nanmax(amplitude)
        else:
            v_max = max_data_value

        # set the contour levels belonging to this subplot
        if min_data_value is None and max_data_value is None:
            # If both limits where not given, assume that we want matplotlib decide on
            # the limits so set levels=None
            levels = None
        else:
            levels = np.linspace(v_min, v_max, number_of_contour_levels + 1)

        # Finally, create the contour plot wit the RAO magnitude
        if use_contourf:
            cs = ax.contourf(
                data_x_2d,
                data_y_2d,
                amplitude,
                cmap=color_map,
                levels=levels,
                zorder=zorder,
            )
        else:
            ls = LightSource(azdeg=315, altdeg=30)
            vert_eg = 1
            L_boundaries = (0, self.Lx, 0, self.Ly)
            rgb = ls.shade(
                amplitude.T, cmap=cm.ocean, vert_exag=vert_eg, blend_mode="hsv"
            )
            cs = plt.imshow(
                rgb,
                origin="lower",
                extent=L_boundaries,
                animated=True,
                vmin=v_min,
                vmax=v_max,
            )

        # if levels is not None:
        #    cs.cmap.set_under("k")
        #    cs.cmap.set_over("k")
        #    cs.set_clim(v_min, v_max)

        # set the axis labels
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        if r_axis_type == "period":
            label_values = ax.get_yticks()
            label_strings = []
            for label in label_values:
                try:
                    label_strings.append(f"{1 / float(label):.1f}")
                except (ValueError, ZeroDivisionError):
                    pass

            ax.set_yticks(label_values)
            ax.set_yticklabels(label_strings)

        # create a contour bar
        cbar = fig.colorbar(cs, ax=ax)
        cbar.ax.set_ylabel("{} [{}]".format("Wave height", "m"))

        if add_hs_estimate:
            hs_estimate = 4 * amplitude.std()
            plt.figtext(
                x_hs_label,
                y_hs_label,
                f"Hs={hs_estimate:.2f} m",
                fontsize=title_font_size,
                verticalalignment="top",
            )

        if plot_title is not None:
            plt.figtext(
                x_plot_title,
                y_plot_title,
                plot_title,
                fontsize=title_font_size,
                verticalalignment="top",
            )
            window_title = plot_title
            fig.canvas.set_window_title(window_title)

        return fig, ax

    @staticmethod
    def _update_plot(
        frame_index=-1,
        self=None,
        fig=None,
        ax=None,
        changed_list=list(),
        min_data_value=0,
        max_data_value=None,
        number_of_contour_levels=10,
        color_map=cc.m_coolwarm,
        use_contourf=True,
        add_hs_estimate=True,
        title_font_size=10,
        x_time_label=0.05,
        y_time_label=0.95,
        x_hs_label=0.05,
        y_hs_label=0.92,
        zorder=0,
    ):
        if frame_index < 0:
            changed_list = list()
        else:
            clean_up_artists(axis=ax, artist_list=changed_list)
            self.wave1D.propagate_wave()
        self.calculate_wave_surface()
        amplitude = self.amplitude
        data_x_2d = self.xy_mesh[0]
        data_y_2d = self.xy_mesh[1]

        if min_data_value is None:
            v_min = np.nanmin(amplitude)
        else:
            v_min = min_data_value

        if max_data_value is None:
            v_max = np.nanmax(amplitude)
        else:
            v_max = max_data_value

        if min_data_value is not None and max_data_value is not None:
            # set the contour levels belonging to this subplot
            levels = np.linspace(
                v_min, v_max, number_of_contour_levels + 1, endpoint=True
            )
        else:
            levels = None

        # finally create the contour plot wit the RAO magnitude
        if use_contourf:
            cs = ax.contourf(
                data_x_2d,
                data_y_2d,
                amplitude,
                cmap=color_map,
                levels=levels,
                zorder=zorder,
            )
        else:
            ls = LightSource(azdeg=315, altdeg=30)
            vert_eg = 1
            L_boundaries = (0, self.Lx, 0, self.Ly)
            rgb = ls.shade(
                amplitude.T, cmap=cm.ocean, vert_exag=vert_eg, blend_mode="hsv"
            )
            cs = plt.imshow(
                rgb,
                origin="lower",
                extent=L_boundaries,
                animated=True,
                vmin=v_min,
                vmax=v_max,
            )

        changed_list.append(cs)

        # if levels is not None:
        #    cs.cmap.set_under("k")
        #    cs.cmap.set_over("k")
        #    cs.set_clim(v_min, v_max)

        if add_hs_estimate:
            hs_estimate = 4 * amplitude.std()
            txt = plt.figtext(
                x_hs_label,
                y_hs_label,
                f"Hs={hs_estimate:.6f} m",
                fontsize=title_font_size,
                verticalalignment="top",
            )
            changed_list.append(txt)

        if frame_index < 0:
            # create a contour bar
            cbar = fig.colorbar(cs, ax=ax)
            cbar.ax.set_ylabel("{} [{}]".format("Wave height", "m"))

        time_label = f"{self.wave1D.time_delta}"
        time_txt = plt.figtext(
            x_time_label,
            y_time_label,
            time_label,
            fontsize=title_font_size,
            verticalalignment="top",
        )
        changed_list.append(time_txt)

        return changed_list

    def animate_wave(
        self,
        figsize=None,
        plot_title=None,
        x_plot_title=0.05,
        y_plot_title=0.99,
        min_data_value=None,
        max_data_value=None,
        number_of_contour_levels=10,
        title_font_size=10,
        x_time_label=0.05,
        y_time_label=0.95,
        x_hs_label=0.05,
        y_hs_label=0.92,
        add_hs_estimate=True,
        color_map=cc.m_coolwarm,
        zorder=0,
        use_contourf=True,
        interval=1,
        title_horizontal_algnment="center",
    ):
        """Create a plot of the current wave

        Parameters
        ----------
        figsize: tuple
            x and y size of the total figure. Default is None, so figure size is not
            imposed
        plot_title: str
            Title to put in th figure. If None, use the titles found in the liftdyn file
        x_plot_title: float, optional
            x position (as a fraction of the sub plot index_title_axis). Default = 0.0
        y_plot_title: float, optional
            y position (as a fraction of the sub plot index_title_axis). Default = 1.1
        delta_y_titles: float, optional
            spacing between the lines of thet title lines. Default = 0.05
        max_title_lines: int or None
            Maxinmum number of title lines to plot. Default = None, ie. plot all
        title_font_size: int
            Size of the title font
        r_axis_lim: tuple or None
            The limits of the x axis. Default = None, so no limits are imposed
        zorder:  int, optional
            Position of the contour plot. Default = 0, meaning that the contours are
            placed at the bottom and the grid and labels are visible
        use_contourf: bool, optional
            If true, use contourf to make the contour plot. Slower. Default = False

        Returns
        -------
        tuple (fig, axis)
            Handle to the figure and the axis
        """
        if figsize is not None:
            size = figsize
        else:
            size = None

        fig, ax = plt.subplots(figsize=size)
        ax.set_aspect("equal", "datalim")

        x_label = "X [m]"
        y_label = "Y [m]"
        # data_x_2d = self.xy_mesh[0]
        # data_y_2d = self.xy_mesh[1]
        amplitude = self.amplitude

        if min_data_value is None:
            v_min = np.nanmin(amplitude)
        else:
            v_min = min_data_value

        if max_data_value is None:
            v_max = np.nanmax(amplitude)
        else:
            v_max = max_data_value

        # set the axis labels
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        if plot_title is not None:
            plt.figtext(
                x_plot_title,
                y_plot_title,
                plot_title,
                fontsize=title_font_size,
                verticalalignment="top",
                ha=title_horizontal_algnment,
            )
            window_title = plot_title
            fig.canvas.set_window_title(window_title)

        # set the contour levels belonging to this subplot
        if min_data_value is None and max_data_value is None:
            # if both limits where not given, assume that we want matplotlib to decide
            # on the limits so set levels to None
            levels = None
        else:
            levels = np.linspace(v_min, v_max, number_of_contour_levels + 1)

        changed_artists = list()

        # Initial call to _update_plot to put of the axis.
        # Return changed_artists with a list of all items which need to be deleted
        # every frame
        changed_artists = self._update_plot(
            self=self,
            fig=fig,
            ax=ax,
            changed_list=changed_artists,
            min_data_value=min_data_value,
            max_data_value=max_data_value,
            number_of_contour_levels=levels,
            color_map=color_map,
            use_contourf=use_contourf,
            add_hs_estimate=add_hs_estimate,
            title_font_size=title_font_size,
            x_time_label=x_time_label,
            y_time_label=y_time_label,
            x_hs_label=x_hs_label,
            y_hs_label=y_hs_label,
        )

        # the fagrs list below must exactly match the arguments of the _update_plot
        # function, except # for the first argument which is the frame_index
        ani = animation.FuncAnimation(
            fig,
            self._update_plot,
            frames=int(self.wave1D.nt_samples),
            fargs=(
                self,
                fig,
                ax,
                changed_artists,
                min_data_value,
                max_data_value,
                number_of_contour_levels,
                color_map,
                use_contourf,
                add_hs_estimate,
                title_font_size,
                x_time_label,
                y_time_label,
                x_hs_label,
                y_hs_label,
                zorder,
            ),
            interval=interval,
            blit=False,
            repeat=True,
        )

        return ani

    def plot_spectrum(
        self,
        figsize=(12, 6),
        r_axis_type="wave_number",
        plot_title=None,
        x_plot_title=0.05,
        y_plot_title=0.99,
        min_data_value=0,
        max_data_value=None,
        number_of_contour_levels=10,
        title_font_size=10,
        r_axis_lim=None,
        polar_projection=False,
        shift_origin=True,
        color_map=cc.m_rainbow,
        color_map_phase=cc.m_coolwarm,
        zorder=0,
        use_contourf=False,
        r_label_position=180,
        kx_min=None,
        kx_max=None,
        ky_min=None,
        ky_max=None,
    ):
        """Create a polar plot of the current spectral data

        Parameters
        ----------
        figsize: tuple
            x and y size of the total figure. Default is None, so figure size is not
            imposed
        r_axis_type: {"wave_number", "wave_length"}, optional
            quantity to use along the radial axis. Either wave_number [rad/m] or
            wave_length [m]. Default = "wave_number"
        plot_title: str
            Title to put in th figure. If None, don't add a title
        x_plot_title: float, optional
            x position (as a fraction of the sub plot index_title_axis). Default = 0.0
        y_plot_title: float, optional
            y position (as a fraction of the sub plot index_title_axis). Default = 1.1
        title_font_size: int
            Size of the title font
        r_axis_lim: tuple or None
            The limits of the x axis. Default = None, so no limits are imposed
        zorder:  int, optional
            Position of the contour plot. Default = 0, meaning that the contours are
            placed at the bottom and the grid and labels are visible
        r_label_position: float, optional
            Angle along which we position the labels of the r-axis
        use_contourf: bool, optional
            Use the slower contourf to create a contour plot. Default = False, i.e.,
            imshow is used
        kx_min: float, optional
            Minimum wave vector node in x direction for cartesian plot. Default = None
        kx_max: float, optional
            Maximum wave vector node in x direction for cartesian plot. Default = None
        ky_min: float, optional
            Minimum wave vector node in y direction for cartesian plot. Default = None
        ky_max: float, optional
            Maximum wave vector node in y direction for cartesian plot. Default = None

        Returns
        -------
        tuple (fig, axis)
            Handle to the figure and the axis
        """

        if figsize is not None:
            size = figsize
        else:
            size = None

        if polar_projection:
            fig, axis = plt.subplots(
                nrows=1, ncols=2, figsize=size, subplot_kw=dict(projection="polar")
            )
            for ax in axis:
                ax.set_theta_zero_location("N")
                ax.set_theta_direction(-1)
        else:
            fig, axis = plt.subplots(nrows=1, ncols=2, figsize=size)
            for ax in axis:
                ax.set_aspect("equal", "datalim")

        if r_axis_lim is not None:
            # the radial limits are but on the ylim
            ax.set_ylim(r_axis_lim)

        if r_axis_type == "wave_number":
            x_label = "Wave number [rad/m]"
            y_label = ""
        elif r_axis_type == "wave_length":
            x_label = "Wave length [m]"
            y_label = ""
        else:
            raise AssertionError(
                "r_axis_type can only be 'wave_number' or 'wave_length'. Found {}"
                "".format(r_axis_type)
            )

        if polar_projection:
            data_x_2d = self.k_polar_mesh[0]
            data_y_2d = self.k_polar_mesh[1]
            psd_2d = abs(self.E_wave_density_polar)
            ang_2d = np.angle(self.E_wave_complex_amplitudes)
        else:
            x_label = "k_x [rad/m]"
            y_label = "k_y [rad/m]"
            if self.wave1D.mirror or shift_origin:
                data_x_2d = fftshift(self.k_cartesian_mesh[0])
                data_y_2d = fftshift(self.k_cartesian_mesh[1])
                psd_2d = fftshift(abs(self.E_wave_density_polar))
                ang_2d = fftshift(np.angle(self.E_wave_complex_amplitudes))
            else:
                data_x_2d = self.k_cartesian_mesh[0]
                data_y_2d = self.k_cartesian_mesh[1]
                psd_2d = abs(self.E_wave_density_polar)
                ang_2d = np.angle(self.E_wave_complex_amplitudes)

        if min_data_value is None:
            v_min = np.nanmin(psd_2d)
        else:
            v_min = min_data_value

        if max_data_value is None:
            v_max = np.nanmax(psd_2d)
        else:
            v_max = max_data_value

        # set the contour levels belonging to this subplot
        if min_data_value is None and min_data_value is None:
            # if both limits where not given, assume that we want matplotlib to decide
            # on the limits so set levels=None
            levels = None
        else:
            levels = np.linspace(v_min, v_max, number_of_contour_levels + 1)

        # finally create the contour plot wit the RAO magnitude
        if use_contourf:
            cs = axis[0].contourf(
                data_x_2d,
                data_y_2d,
                psd_2d,
                cmap=color_map,
                levels=levels,
                zorder=zorder,
            )
            cs2 = axis[1].contourf(
                data_x_2d,
                data_y_2d,
                ang_2d,
                cmap=color_map_phase,
                levels=None,
                zorder=zorder,
            )
        else:
            cs = axis[0].imshow(psd_2d.T, origin="lower", cmap=color_map)
            cs2 = axis[1].imshow(ang_2d.T, origin="lower", cmap=color_map_phase)

        # if levels is not None:
        # if levels is not None:
        #    cs.cmap.set_under("k")
        #    cs.cmap.set_over("k")
        #    cs.set_clim(v_min, v_max)
        # set the axis labels

        if r_axis_type == "wave_length":
            label_values = ax.get_yticks()
            label_strings = []
            for label in label_values:
                try:
                    label_strings.append(f"{2 * np.pi / float(label):.1f}")
                except (ValueError, ZeroDivisionError):
                    pass

            ax.set_yticks(label_values)
            ax.set_yticklabels(label_strings)

        # set the axis labels
        for ax in axis:
            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)

            if polar_projection:
                ax.set_rlabel_position(r_label_position)

                if r_axis_lim is not None:
                    # the radial limits are but on the ylim
                    ax.set_ylim(r_axis_lim)
            else:
                set_limits(axis=ax, v_min=kx_min, v_max=kx_max, direction="x")
                set_limits(axis=ax, v_min=ky_min, v_max=ky_max, direction="y")
                ax.set_aspect("equal", "datalim")

        # create a contour bar
        cbar = fig.colorbar(cs, ax=axis[0])
        cbar.ax.set_ylabel("{} [{}]".format("PSD", "m3"))

        cbar2 = fig.colorbar(cs2, ax=axis[1])
        cbar2.ax.set_ylabel("{} [{}]".format("Phase", "rad"))

        if plot_title is not None:
            plt.figtext(
                x_plot_title,
                y_plot_title,
                plot_title,
                fontsize=title_font_size,
                verticalalignment="top",
            )
            window_title = plot_title
            fig.canvas.set_window_title(window_title)

        return fig, axis


class Wave1D:
    r"""
    A class for linearized solutions of 1D wave vector equation

    .. math ::

        \eta(x_p, t) = \sum_i^{n_k} a_i \exp(j (k_i x_p - \omega_i t + \phi_i))

    where the amplitudes :math:`a_i` follow from the Jonswap or Gauss power spectral
    density function. See the top of the module for more information

    Parameters
    ----------
    xmin: float, optional
        Start of spatial domain. Default = 0
    Lx: float, optional
        Length of spatial domain. Default = 100
    t_start: float, optional
        Start of temporal domain. Default = 0
    t_length: float, optional
        Length of temporal domain. Default = 100
    nt_samples: int, optional
        Number of time samples. Default = 256
    nx_points: float, optional
        Number of space domain samples in x direction. Default = 256
    E_limit_low: float, optional
        Fraction of energy to cut at the low end of the spectrum.  Default = 0.01
    E_limit_high: float, optional
        Fraction of energy to keep at the high end of the spectrum.  Default = 0.9
        (so 0.1 is cut off at the high end)
    kx_min: float, optional
        Minimal value of the wave vector domain. Default = 0
    kx_max: float, optional
        Maximum value of the wave vector domain. Default = pi/2
    delta_kx: float, optional
        Spacing in wave vector domain.
        Default = 0.06283 ((kx_max - kx_min) / n_kx_nodes)
    n_kx_nodes: int, optional
        Number of nodes in the wave vector domain. Default = 256
    n_bins_equal_energy: int, optional
        Number of nodes in the wave vector domain in case we are in the EqualEnergyBins
        mode. See `wave_selection` below. Default value = 100
    Hs: float, optional
        Significant wave height [m]. Default = 3 m
    Tp: float, optional
        Peak period [s]. Default = 10 s
    gamma: float, optional
        Peakness factor in case Jonswap spectrum is used. Default = 3.3
    sigma: float, optional
        Width of spectrum in case Gauss spectrum  with *spectral_version='dnv'* is used.
        Default = 0.0625
    random_phase: bool, optional
        Use random phase with each run. Default = False
    spectrum_type: {"jonswap", "gauss"}, optional
        Type of spectrum used. Default = "jonswap"
    spectral_version: {"sim", "dnv"}, optional
        Version of wave spectrum used. Default = "sim". For the Jonswap spectrum there
        is virtually no difference between the "sim" and "dnv" version. For the Gauss
        spectrum the "sim" version has a spectral width related to Tp which is fixed,
        wheras the "dnv" version has a width based on the *sigma* input argument.
    gravity0: float, optional
        Gravitation constant. Default = g0 = 9.81
    wave_construction: {"FFT", "DFTpolar", "DFTcartesian"}
        Method how the wave field is constructed from the spectrum. Default = "FFT". The
        options are

        * *DFTpolar*: A DFT (see top of module) is used to calculate the wave field.
          Slow, however, the location  of the wave vectors is not restricted.
          Therefore, a smart selection of the wave nodes can be made. See
          *wave_selection*  for the possibilities.
        * *FFT* : The wave is constructed from the wave vector nodes using a Fast
          Fourier Transfor (FFT). This implies that the spectrum nodal poits are
          restricted: the number of wave nodes equals the number of spatial points x,
          and more over, the complex amplitude must obey the symmetry
          rule :math:`a(k)=a^*(-k)`. The implies that we can not make a wave selection.
          However, the FFT algorithm is so much faster than the DFT that we can still
          obtain faster calculations time using the FFT based on all wave nodes.
          The FFT wave construction is therefore recommended.
        * *DFTcartesian*: This choice is for validation purpose only. It assumes the
          same symmetric spectrum as used for the FFT option but then still the slow
          DFT  is used to calculate the wave field.
    wave_selection: {"All", "EqualEnergyBins", "Subrange"}
        For the DFTPolar wave construction mode we can make a selection of wave
        components in order to speed up the wave calculation. Three choices are possible

        * All: No selection is made so all the wave vectors as defined in the kx_nodes
          domain are used
        * Subrange: A Subrange of wave vectors based on the E_limit_low and E_limit_high
          values is made
        * EqualEnergyBins: The energy per bin is assumed equal
    use_subrange_energy_limits : bool
        If true the wave selection in the EqualEnergyBins mode is also limited by the
        Subrange settings. Default = True
    sample_every : int
        Make a wave selection by taking every 'sample_every' point in the wave vector
        domain. Only applicable when the wave_selection modes is *Subrange*

    Attributes
    ----------
    kx_nodes : ndarray
       Wave vector nodes of size :math:`n_k` used to build the spectrum and the wave
       field. The values depend on the wave_selection mode and wave_construction type.
       In case FFT is used, the kx_nodes are symmetrical around 0 to be able to define
       the symmetric spectrum.
    xpoints : ndarray
       Spatial position array of size :math:`n_x`
    amplitude : ndarray
       Wave height along x-direction at time t of size :math:`n_x`
    phase : ndarray
       Random phase array of size :math:`n_k`
    omega_dispersion : ndarray
       Angular frequency per wave node of size :math:`n_k` following from the deep water
       dispersion relation
    spectrumK : ndarray
       Power spectral density of the wave spectrum along the wave vector nodes k of
       size :math:`n_k`
    spectrumW : ndarray
       Power spectral density of the wave spectrum along the angular frequencies of size
       :math:`n_k`. Note that since the angular frequency are coupled to the wave vector
       nodes via the dispesion relation, the bin spacing of this vector is not
       equidistant
    complex_amplitudes : ndarray
        The complex amplitudes following from the power spectral density spectrumK and
        the phase vector as S_k * exp(j phase)
    delta_x :  float
        Spatial resolution of the x-domain
    kx_nyquist :  float
        Maximum wave vector following from the spatial resolution
        *delta_x* as  pi / delta_x

    Examples
    --------

    A wave can be setup a wave with all the default settings  and plot the spectrum and
    wave at time 0 as follows

    >>> wave_dft = Wave1D()
    >>> fig, axis = wave_dft.plot_spectrum()
    >>> fig, axis = wave_dft.plot_wave()

    This will create the wave based on the DFT, hence it is slow. A wave construction
    based on the FFT can be picked as

    >>> wave_fft = Wave1D()

    An animation of the wave can be made as follows

    >>> movie = wave_fft.animate_wave()

    This will infinitely simulate the wave. We can put and finit time by setting the
    t_length to the wave

    >>> wave_fft.reset_time(t_length=100)

    For more examples, please have a look at the notebook linked via the Example section
    example_wave_fields
    """

    def __init__(
        self,
        name=None,
        xmin=0,
        Lx=None,
        xmax=None,
        t_start=0,
        delta_t=1,
        t_length=None,
        nt_samples=1000000,
        nx_points=512,
        E_limit_low=0.01,
        E_limit_high=0.90,
        kx_min=0.0,
        kx_max=1.5708,
        delta_kx=0.06283,
        n_kx_nodes=512,
        n_bins_equal_energy=64,
        use_subrange_energy_limits=False,
        sample_every=1,
        Hs=3.0,
        Tp=10.0,
        gamma=3.3,
        sigma=0.0625,
        random_phase=False,
        wave_construction="FFT",
        wave_selection="All",
        lock_nodes_to_wave_one=False,
        spectrum_type="jonswap",
        spectral_version="sim",
        gravity0=g0,
    ):
        self.spectrum_type = spectrum_type
        self.spectral_version = spectral_version

        logger.info(f"Initialise {self.spectrum_type} 1D wave field")

        if name is None:
            self.name = "_".join(
                ["wave", spectrum_type, wave_construction, wave_selection, "1d"]
            )
        else:
            self.name = name

        # current time settings
        self.time = 0
        self.time_delta = pd.Timedelta(self.time, unit="s")
        self.t_index = 0
        self.repeat_movie = False
        self.playing_movie = False

        self.xmin = xmin
        if xmax is not None:
            self.Lx = xmax - self.xmin
            if Lx is not None:
                logger.warning(
                    "Both maximum x value xmax={} and length of domain Lx={} are "
                    "defined. Taking xmax leading so set Lx to {}"
                    "".format(xmax, Lx, self.Lx)
                )
        else:
            if Lx is not None:
                self.Lx = Lx
            else:
                self.Lx = 1000
        self.xmax = self.xmin + self.Lx

        self.t_start = None
        self.delta_t = None
        self.nt_samples = None
        self.t_length = None
        self.t_end = None
        self.reset_time(
            t_length=t_length, t_start=t_start, nt_samples=nt_samples, delta_t=delta_t
        )

        self.nx_points = nx_points
        self.nt_samples = nt_samples

        self.Hs = Hs
        self.Tp = Tp

        self.gravity0 = gravity0

        # depending on the wave type we need to define gamma or sigma (for Jonswap or
        # Gauss, resp.)
        self.gamma = gamma
        self.sigma = sigma

        self.E_limit_low = E_limit_low
        self.E_limit_high = E_limit_high
        self.kx_min = kx_min
        self.kx_max = kx_max
        self.delta_kx = delta_kx
        self.delta_kx_non_uni = None
        self.n_kx_nodes = n_kx_nodes

        self.random_phase = random_phase

        # Select which wave component are selected. Choices are: Subrange,
        # EqualEnergyBins, All
        self.wave_selection = wave_selection

        # set the wave construction
        self.wave_construction = None
        self.mirror = False
        self.set_wave_construction(wave_construction)

        if self.wave_construction == "FFT" and self.wave_selection != "All":
            raise ValueError(
                "Can not combine 'FFT' wave_construction method with a "
                "wave_selecting other than 'All'"
            )

        self.lock_nodes_to_wave_one = lock_nodes_to_wave_one

        # use seed to update the random phase if required
        self.seed = 1
        self.update_phase = True

        self.pick_single_wave = False
        self.picked_wave_index = 0

        self.n_bins_equal_energy = n_bins_equal_energy
        self.use_subrange_energy_limits = use_subrange_energy_limits
        self.sample_every = sample_every

        self.xpoints = None
        self.phase = None
        self.kx = None

        self.amplitude = None
        self.delta_x = None
        self.kx_nyquist = None

        self.eta = 0

        self.ik_low = None
        self.ik_high = None
        self.ik_peak = None
        self.k_low = None
        self.k_high = None
        self.k_peak = None
        self.a_peakK = None
        self.varianceK = None
        self.iW_low = None
        self.iW_high = None
        self.iW_peak = None
        self.W_low = None
        self.W_high = None
        self.W_peak = None
        self.a_peakW = None
        self.varianceW = None
        self.spectrumK = None

        self.omega_dispersion = None
        self.delta_omega = None

        self.complex_amplitudes = None

        # the minimum and maximum wave
        self.global_wave_extremes = [0, 0]

        # add all the properties of the plot of the current wave
        self.plot = PlotProperties()

        self.update_x_k_t_sample_space()
        self.calculate_spectra_modulus()

    def reset_time(self, t_length=None, t_start=0, nt_samples=10000000, delta_t=1):
        """Reset all time properties and allow to recalculate"""
        self.t_start = t_start
        self.delta_t = delta_t
        self.nt_samples = nt_samples
        if t_length is None:
            self.t_length = self.nt_samples * self.delta_t
        else:
            self.t_length = t_length
            self.nt_samples = int(self.t_length / self.delta_t)
        self.t_end = self.t_start + self.t_length
        self.time = self.t_start
        self.t_index = 0
        self.time_delta = pd.Timedelta(self.time, unit="s")

    def make_report(self):
        """Make report of settings for this wave"""

        frm = "{:40s} : {}"

        logger.info(frm.format("Name", self.name))
        logger.info("----------- Domain settings --------")
        logger.info(frm.format("Start x [m]", self.xmin))
        logger.info(frm.format("End x [m]", self.xmax))
        logger.info(frm.format("Length x [m]", self.Lx))
        logger.info("----------- Time settings --------")
        logger.info(frm.format("Start t [s]", self.t_start))
        logger.info(frm.format("End t [s]", self.t_end))
        logger.info(frm.format("Length t [s]", self.t_length))
        logger.info(frm.format("Delta t [s]", self.delta_t))

        logger.info("----------- Numerical resolutions --------")
        logger.info(frm.format("Number x - nodes", self.xpoints.size))
        logger.info(frm.format("Number kx - nodes", self.kx_nodes.size))
        logger.info(frm.format("Delta x [m]", self.delta_x))
        dkx = np.diff(self.kx_nodes)
        logger.info(frm.format("Delta kx min [rad/m]", dkx.min()))
        logger.info(frm.format("Delta kx max [rad/m]", dkx.max()))
        logger.info(frm.format("Delta kx first [rad/m]", self.delta_kx))
        logger.info(frm.format("k-min [rad/m]", self.kx_nodes.min()))
        logger.info(frm.format("k-max [rad/m]", self.kx_nodes.max()))
        logger.info(frm.format("k-nyq [rad/m]", self.kx_nyquist))
        logger.info(frm.format("L_max [m]", 2 * np.pi / dkx[0]))

        logger.info("----------- Numerical methodds --------")
        logger.info(frm.format("Selection method", self.wave_selection))
        logger.info(frm.format("Construction method", self.wave_construction))

    def next_time(self):
        """Increase the time"""
        self.time += self.delta_t
        self.time_delta = pd.Timedelta(self.time, unit="s")
        self.t_index += 1
        if self.time > self.t_end:
            if self.repeat_movie:
                self.time = self.t_start
                self.t_index = 0
            else:
                self.playing_movie = False

    def propagate_wave(self):
        """Increase the time and recalculate the surface"""

        self.next_time()
        self.calculate_wave_surface()

    def animate_wave(
        self,
        x_min=None,
        x_max=None,
        y_min=None,
        y_max=None,
        figsize=(12, 4),
        interval=25,
        x_time_label=0.01,
        y_time_label=0.93,
        plot_title=None,
        x_plot_title=0.5,
        y_plot_title=0.97,
        title_font_size=12,
        title_horizontal_algnment="center",
    ):
        """Animate the 1D wave vs space and wave vs time

        Parameters
        ----------
        fig: :obj:`figure`
            Figure returned by previous call can be update
        ax: :obj:`Axes`
            Axes returned by previous call can be updated to add more waves
        x_min: float, optional
            Start of the x-range. Default = None
        x_max: float, optional
            End of the x-range. Default = None
        y_min: float, optional
            Start of the y-range. Default = None
        y_max: float, optional
            End of the y-range. Default = None
        figsize: tuple, optional
            Tuple with (with, height) in inches. Default = None
        interval: int, optional
            Time in ms between each frame. Default = 25
        x_time_label: float, optional
            X position relative to the current graph where to put the time label
        y_time_label: float, optional
            Y Position relative to the current graph where to put the time label

        Returns
        -------
        tuple
            (fig, ax) reference to the figure and its axis
        """

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

        self.calculate_wave_surface()
        (line,) = ax.plot(self.xpoints, self.amplitude)
        ax.set_xlabel("x position [m]")
        ax.set_ylabel("wave amplitude [m]")

        time_label_string = "{:02d}:{:02d}:{:02d}"

        time_text = ax.text(
            x_time_label,
            y_time_label,
            time_label_string.format(
                self.time_delta.components.hours,
                self.time_delta.components.minutes,
                self.time_delta.components.seconds,
            ),
            transform=ax.transAxes,
        )

        if plot_title is not None:
            plt.figtext(
                x_plot_title,
                y_plot_title,
                plot_title,
                fontsize=title_font_size,
                verticalalignment="top",
                ha=title_horizontal_algnment,
            )
            window_title = plot_title
            fig.canvas.set_window_title(window_title)

        set_limits(axis=ax, v_min=y_min, v_max=y_max, direction="y")

        self.playing_movie = True

        def next_time():
            # An iterator to proceed to the next wave and yield the index only if we
            # want to keep playing.
            # Obsolete, I have now added the proceed wave to the animate function and
            # past nt_samples
            logger.debug(f"Updating wave for time {self.time}")
            self.propagate_wave()
            if self.playing_movie:
                yield self.t_index

        def animate(i):
            logger.debug(
                "Plotting wave for time {:10.1f} s until {:10.1f} s :  {}".format(
                    self.time, self.t_end, self.time_delta
                )
            )
            self.propagate_wave()
            line.set_ydata(self.amplitude)
            time_text.set_text(
                time_label_string.format(
                    self.time_delta.components.hours,
                    self.time_delta.components.minutes,
                    self.time_delta.components.seconds,
                )
            )
            return line, time_text

        def init():
            # Initialise the wave plot by first calculating the current wave  height
            # and than set the data
            line.set_ydata(self.amplitude)
            return (line,)

        ani = animation.FuncAnimation(
            fig,
            animate,
            self.nt_samples,
            init_func=init,
            interval=interval,
            blit=False,
            repeat=True,
        )
        return ani

    @staticmethod
    def limit_markers(
        ax, text, a_x, a_y, s_x=0, s_y=0, symbol="vy", arrow_color="black", fontsize=8
    ):
        ax.plot([a_x], [a_y], symbol)
        ax.annotate(
            text,
            xy=(a_x, a_y),
            xytext=(s_x, s_y),
            textcoords="offset points",
            horizontalalignment="left",
            va="top",
            arrowprops=dict(width=2, headwidth=5, shrink=0.1, color=arrow_color),
            fontsize=fontsize,
        )

    def plot_spectrum(
        self,
        fig=None,
        axis=None,
        x_min=None,
        x_max=None,
        y_min=None,
        y_max=None,
        figsize=None,
        add_limit_markers=False,
        add_hs_estimate=False,
        add_n_points=True,
        y_n_points_label=0.8,
        line_markers="o",
        markersize=4,
        linestyle="-",
        linecolor=None,
        plot_title=None,
        x_plot_title=0.5,
        y_plot_title=0.97,
        title_font_size=12,
        title_horizontal_algnment="center",
    ):
        """Plot the spectrum in k-spaco

        Parameters
        ----------
        fig: :obj:`Figure`, optional
            Reference to the figure to add the plot to
        axis: :obj:`Axis`, optional
            Reference to the axis
        x_min: float, optional
            set minimum x value, Default = None i.e take the default
        x_max: float, optional
            set maximum x value, Default = None i.e take the default
        y_min: float, optional
            set minimum y value, Default = None i.e take the default
        y_max: float, optional
            set maximum y value, Default = None i.e take the default
        figsize: tuple
            (width, height) of Figure. Default = None ie. take the default
        add_limit_markers: bool, optional
            Add markers to indicate the low and high limits of the energy in the
            spectrum
        add_hs_estimate=False, optional
            Add a label indicating the Hs belonging to the energy of the spectrum
        add_n_points: bool, optional
            Add a label with the number of points
        y_n_points_label: bool, optional
            Give the y coordinates  the n_point label, Default = 0.8
        line_markers: str, optional
            Give the marker of the plot. Default = "o". Set to "" to remove the markers
        linestyle: str, optional
            Give the linestyle, Default = "-"
        linecolor: str, optional
            Give the color of the line. Default = None, ie take the default
        markersize: int, optional
            Give the size of the markers. Default = 4

        Returns
        -------
        tuple
            (fig, axis), reference to the create fig and axis

        """
        if fig is None or axis is None:
            fig, axis = plt.subplots(nrows=2, ncols=1, figsize=figsize)
            plt.subplots_adjust(hspace=0.25)
            axis[0].set_ylabel("spectral amplitude [m3]")
            axis[0].set_xlabel("wave vector kx [rad/m]")

            axis[1].set_ylabel("spectral amplitdude [m2s]")
            axis[1].set_xlabel("omega [rad/s]")

        if self.mirror:
            kx_nodes = fftshift(self.kx_nodes)
            omega_disp = fftshift(self.omega_dispersion)
            spectrum_k = fftshift(self.spectrumK)
            spectrum_w = fftshift(self.spectrumW)
        else:
            kx_nodes = self.kx_nodes
            omega_disp = self.omega_dispersion
            spectrum_k = self.spectrumK
            spectrum_w = self.spectrumW

        (line0,) = axis[0].plot(
            kx_nodes,
            spectrum_k,
            linestyle=linestyle,
            marker=line_markers,
            markersize=markersize,
            color=linecolor,
        )

        (line1,) = axis[1].plot(
            omega_disp,
            spectrum_w,
            linestyle=linestyle,
            marker=line_markers,
            markersize=markersize,
            color=linecolor,
        )

        if add_limit_markers:
            self.limit_markers(
                ax=axis[0],
                a_x=self.k_low,
                a_y=self.spectrumK[self.ik_low],
                symbol="vy",
                text=f"E(k<{self.k_low:.2g})/E\n={self.E_limit_low:.2g}",
                s_x=-50,
                s_y=40,
            )
            self.limit_markers(
                ax=axis[0],
                a_x=self.k_high,
                a_y=self.spectrumK[self.ik_high],
                symbol="^g",
                text="E(k>{:.2g})/E\n={:.2g}".format(
                    self.k_high, 1 - self.E_limit_high
                ),
                s_x=10,
                s_y=40,
            )
            self.limit_markers(
                ax=axis[0],
                a_x=self.k_peak,
                a_y=self.spectrumK[self.ik_peak],
                symbol="or",
                text="k={:.2g} rad/m\nL={:.1f} m".format(
                    self.k_peak, 2 * np.pi / self.k_peak
                ),
                s_x=20,
                s_y=-30,
            )

            self.limit_markers(
                ax=axis[1],
                a_x=self.W_low,
                a_y=self.spectrumW[self.iW_low],
                symbol="vy",
                text=f"E(k<{self.W_low:.2g})/E\n={self.E_limit_low:.2g}",
                s_x=-50,
                s_y=40,
            )
            self.limit_markers(
                ax=axis[1],
                a_x=self.W_high,
                a_y=self.spectrumW[self.iW_high],
                symbol="^g",
                text="E(k>{:.2g})/E\n={:.2g}".format(
                    self.W_high, 1 - self.E_limit_high
                ),
                s_x=10,
                s_y=40,
            )
            self.limit_markers(
                ax=axis[1],
                a_x=self.W_peak,
                a_y=self.spectrumW[self.iW_peak],
                symbol="or",
                text="k={:.2g} rad/s\nT={:.1f} s".format(
                    self.W_peak, 2 * np.pi / self.W_peak
                ),
                s_x=20,
                s_y=-30,
            )

        if add_hs_estimate:
            var_k = np.sum(self.spectrumK * self.delta_kx)
            if self.wave_construction in ("FFT", "DFTcartesian"):
                var_k /= 2.0
            hs_estimate_k = 4 * np.sqrt(var_k)
            axis[0].text(
                0.85,
                0.92,
                f"Hs={hs_estimate_k:.1f} m",
                transform=axis[0].transAxes,
                va="top",
            )
            var_w = np.sum(self.spectrumW * self.delta_omega)
            if self.wave_construction in ("FFT", "DFTcartesian"):
                var_w /= 2.0
            hs_estimate_w = 4 * np.sqrt(var_w)
            axis[1].text(
                0.85,
                0.92,
                f"Hs={hs_estimate_w:.1f} m",
                transform=axis[1].transAxes,
            )
        if add_n_points:
            axis[0].text(
                0.85,
                y_n_points_label,
                f"N={self.spectrumK.size:d}",
                transform=axis[0].transAxes,
                va="top",
            )

        # for ax in axis:
        #    set_limits(ax, v_min=x_min, v_max=x_max, axis="x")
        #    set_limits(ax, v_min=y_min, v_max=y_max, axis="y")

        if plot_title is not None:
            plt.figtext(
                x_plot_title,
                y_plot_title,
                plot_title,
                fontsize=title_font_size,
                verticalalignment="top",
                ha=title_horizontal_algnment,
            )
            window_title = plot_title
            fig.canvas.set_window_title(window_title)

        return fig, axis

    def plot_wave(
        self,
        fig=None,
        ax=None,
        x_min=None,
        x_max=None,
        y_min=None,
        y_max=None,
        figsize=(12, 4),
        linestyle="-",
        line_markers=None,
        linecolor=None,
        markersize=5,
        label=None,
        plot_title=None,
        x_plot_title=0.5,
        y_plot_title=0.97,
        title_font_size=12,
        title_horizontal_algnment="center",
    ):
        """Plot the 1D wave vs space and wave vs time

        Parameters
        ----------
        fig: :obj:`figure`
            Figure returned by previous call can be update
        ax: :obj:`Axes`
            Axes returned by previous call can be updated to add more waves
        x_min: float, optional
            Start of the x-range. Default = None
        x_max: float, optional
            End of the x-range. Default = None
        y_min: float, optional
            Start of the y-range. Default = None
        y_max: float, optional
            End of the y-range. Default = None
        figsize: tuple, optional
            Tuple with (with, height) in inches. Default = None
        plot_title: str, optional
            Add a title to the plot. Default = None, ie. don't add a title
        x_plot_title: float, optional
            X Location of the plot title. Default = 0.5 wro the figurek
        y_plot_title: float, optional
            Y Location of the plot title. Default = 0.97 wro the figurek
        title_font_size: int, optional
            Font size of the title

        Returns
        -------
        tuple
            (fig, ax) reference to the figure and its axis
        """

        self.calculate_wave_surface()

        if fig is None or ax is None:
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

        (line1,) = ax.plot(
            self.xpoints,
            self.amplitude,
            linestyle=linestyle,
            marker=line_markers,
            markersize=markersize,
            color=linecolor,
            label=label,
        )
        ax.set_xlabel("x position [m]")
        ax.set_ylabel("wave amplitude [m]")

        set_limits(axis=ax, v_min=x_min, v_max=x_max, direction="x")
        set_limits(axis=ax, v_min=y_min, v_max=y_max, direction="y")

        if plot_title is not None:
            plt.figtext(
                x_plot_title,
                y_plot_title,
                plot_title,
                fontsize=title_font_size,
                verticalalignment="top",
                horizontalalignment=title_horizontal_algnment,
            )
            window_title = plot_title
            fig.canvas.set_window_title(window_title)

        return fig, ax

    def set_wave_construction(self, mode):
        """
        Set the wave how the wave field is constructed from the power spectral density:
        via FFT or DFT

        Parameters
        ----------
        mode: {"FFT", "DFTpolar", "DFTcartesian"}
            Construction type of the wave field. In case FFT or DFTcartesian is chosen,
            the spectrum most be symmetric and therefore mirror must be True

        """
        if self.wave_selection != "All" and mode != "DFTpolar":
            logger.warning(
                "You have set a wave selection but want to use an FFT. Selecting waves "
                "is only possible for DFTpolar. Setting that now"
            )
            mode = "DFTpolar"

        self.wave_construction = mode
        if mode in ("FFT", "DFTcartesian"):
            self.mirror = True
        elif mode == "DFTpolar":
            self.mirror = False
        else:
            raise AssertionError(
                "Wave construction must be FFT, DFTcartesian, or DFTpolar. Found {}"
                "".format(mode)
            )

    def update_x_k_t_sample_space(self):
        """
        After a change of number of x-points or wave vector nodes k has been made, call
        this routine to update all the arrays

        """

        self.xmax = self.xmin + self.Lx

        self.xpoints = np.linspace(self.xmin, self.xmax, self.nx_points, endpoint=True)

        self.amplitude = np.zeros(self.xpoints.shape)

        self.delta_x = self.xpoints[1] - self.xpoints[0]

        logger.debug(
            "XMIN {} XMAX {} nx {}  deltax {}".format(
                self.xmin, self.xmax, self.nx_points, self.delta_x
            )
        )
        self.kx_nyquist = np.pi / self.delta_x

        if self.wave_construction == "DFTpolar":
            # If a DDT is used for the wave field calculation, you can use any amount
            # of wave vectors
            if self.kx_max > self.kx_nyquist:
                logger.warning(
                    "kx_max larger than nyquist belongin to dx={}: clipping kx to {}"
                    "".format(self.delta_x, self.kx_nyquist)
                )
                self.kx_max = self.kx_nyquist

            self.kx_nodes = np.linspace(self.kx_min, self.kx_max, self.n_kx_nodes)

            self.delta_kx = self.kx_nodes[1] - self.kx_nodes[0]
        else:
            # for the FFT the number wave vectors should be equal to the number of
            # x-points the DFT based on cartesian values in this case take the same
            # mesh as FFT but then uses the DFT algorith for comparison
            self.kx_nodes = 2 * np.pi * np.fft.fftfreq(self.nx_points, self.delta_x)
            self.delta_kx = self.kx_nodes[1] - self.kx_nodes[0]

            logger.debug(
                "kx_nodes minx/max {} {}".format(
                    np.amin(self.kx_nodes), np.amax(self.kx_nodes)
                )
            )

        if (
            self.phase is None
            or self.kx_nodes.shape != self.phase.shape
            or self.update_phase
        ):
            self.phase = ms.initialize_phase(self.kx_nodes, self.seed)
            self.update_phase = False

        self.t_end = self.t_start + self.t_length

        self.delta_t = self.t_length / self.nt_samples

    def calculate_omega_dispersion(self):
        """
        Calculate the omega frequency belonging to the wave vectors according to the
        deep water dispersion relation.
        """
        self.omega_dispersion = np.sqrt(self.gravity0 * abs(self.kx_nodes)) * np.where(
            self.kx_nodes >= 0,
            np.ones(self.kx_nodes.shape),
            -np.ones(self.kx_nodes.shape),
        )
        delta_omega = np.diff(self.omega_dispersion)
        self.delta_omega = np.append(delta_omega, [delta_omega[-1]])

    def calculate_spectra_modulus(self):
        """
        This routine calculates the Spectrum in omega and k domain. The spectrum can
        be either a Jonswap or a Gauss spectrum depending on the wave_type argument
        Based on the energy limits, the minimum and maximum frequency/wave vector is
        obtained using the specspecs routine. Based on those limits, you may take a
        selection of wave vectors if the wave_selection variable is set to 'subrange' or
        'EqualEnergyBins'
        """

        # self.phase = ms.initialize_phase(self.kx_nodes,self.seed)

        self.spectrumK = ms.spectrum_wave_k_domain(
            k_waves=self.kx_nodes,
            Hs=self.Hs,
            Tp=self.Tp,
            gamma=self.gamma,
            sigma=self.sigma,
            spectrum_type=self.spectrum_type,
            spectral_version=self.spectral_version,
        )

        (
            self.ik_low,
            self.ik_high,
            self.ik_peak,
            self.k_low,
            self.k_high,
            self.k_peak,
            self.a_peakK,
            self.varianceK,
        ) = ms.specspecs(
            self.kx_nodes,
            self.spectrumK,
            lowlim=self.E_limit_low,
            higlim=self.E_limit_high,
            mirror=self.mirror,
        )

        # Select a sub range of the wave vectors. In case that fft is used, this is
        # not possible.
        if self.wave_construction == "DFTpolar" and self.wave_selection == "Subrange":
            # extra wave vectors based on the range k_low,k_high

            # Create a mask array to select the wave vectors within the subrange k_low
            # k_high
            mask = ms.mask_out_of_range(self.kx_nodes, self.k_low, self.k_high)

            # make a selection of wave vectors: set value equal to zero outside range
            self.kx_nodes = np.extract([mask], [self.kx_nodes])
            self.spectrumK = np.extract([mask], [self.spectrumK])
            self.phase = np.extract([mask], [self.phase])

            # Subsample the wave vectors and frequencies bin in case sample_every is
            # larger than one
            if self.sample_every > 1:
                self.kx_nodes = self.kx_nodes[:: self.sample_every]
                self.spectrumK = self.spectrumK[:: self.sample_every]
                self.phase = self.phase[:: self.sample_every]

        elif (
            self.wave_construction == "DFTpolar"
            and self.wave_selection == "EqualEnergyBins"
        ):
            # If EqualEnergy bins is selected, make a selection of frequency bins such
            # that the energy per interval S*dk remains constant to Ebin (the mean
            # energy per bin based on the total energy and number of bins
            self.Ebin = self.varianceK / self.n_bins_equal_energy
            kx_nodes = []
            kk = self.kx_min
            if self.use_subrange_energy_limits:
                kk = self.k_low
            kx_nodes.append(kk)
            logger.debug("Start solving the euqal energy bins...")
            n_trail_max = 10
            while kk < self.kx_max:
                logger.debug(f"Solving bin for kk = {kk}")
                # Calculate the next wave vector such that S*dk (energy in this bin)
                # equals Ebin by solving the integral int_k0^knew Sdk. kk+delta_kx is
                # just a first guess
                delta_kx = self.delta_kx
                n_trail = n_trail_max
                found_solution = False
                while not found_solution and n_trail > 0:
                    ans = sp.optimize.fsolve(
                        _energy_deficit,
                        kk + delta_kx,
                        args=(
                            kk,
                            self.Ebin,
                            self.Hs,
                            self.Tp,
                            self.gamma,
                            self.spectral_version,
                            self.spectrum_type,
                            self.sigma,
                        ),
                    )
                    if ans[0] > 0:
                        found_solution = True
                        kk = ans[0]
                    else:
                        delta_kx /= 2
                        n_trail -= 1
                        # we could not find a solution. Try again with a smaller offset
                        logger.debug(
                            "Failed to find a solution for kk. Try again for smaller"
                            " delta kx {} {}".format(delta_kx, n_trail)
                        )
                if n_trail < 0:
                    logger.warning(
                        "Tried {} times without succes. Stop solving, "
                        "hope for the best".format(n_trail)
                    )
                    break
                # check if kk is within subrange is requested and then add it to the
                # list
                if kk < self.kx_max and not (
                    self.use_subrange_energy_limits
                    and (kk < self.k_low or kk > self.k_high)
                ):
                    logger.debug(
                        "Storing bin kk = {} with k_low = {} k_high ={}"
                        "".format(kk, self.k_low, self.k_high)
                    )
                    kx_nodes.append(kk)
            logger.debug("Done")

            # turn the create kx_node list into a nparray
            if self.lock_nodes_to_wave_one:
                # If lock_nodes_to_wave_one is true, then the k values calculated above
                # are all # locked to the wave vectors of the full wave domain by
                # finding the nearest wave value
                mask = np.full(self.kx_nodes.shape, False, dtype=np.bool)
                for kk in kx_nodes:
                    kinx = find_idx_nearest_val(self.kx_nodes, kk)
                    mask[kinx] = True

                # select the  kx and phase at the nodes
                self.kx_nodes = np.extract(mask, self.kx_nodes)
                self.phase = np.extract(mask, self.phase)
            else:
                # in case the nodes do not have to be locked to the first wave,
                # convert the created list of wave vectors into a numpy array, calculate
                # a new phase and the corresponding modulus
                self.kx_nodes = np.array(kx_nodes)
                self.phase = ms.initialize_phase(self.kx_nodes, self.seed)

            # Based on the locked or non-locked kx_nodes with fixed and non-fixed
            # phases, calculate the spectrumK
            self.spectrumK = ms.spectrum_wave_k_domain(
                k_waves=self.kx_nodes,
                Hs=self.Hs,
                Tp=self.Tp,
                gamma=self.gamma,
                sigma=self.sigma,
                spectrum_type=self.spectrum_type,
                spectral_version=self.spectral_version,
            )

        elif self.wave_selection == "OneWave":
            logger.debug(f"selecting wave {self.picked_wave_index}")
            # pick one wave only
            k = self.picked_wave_index
            if not self.wave_construction == "FFT":
                # only copy the single wave of the picked wave vector for the DFT
                # algorithms
                self.kx_nodes = np.array([self.kx_nodes[k]])
            else:
                # keep all the wave vectors except set the value not equal to the picked
                # wave to zero
                mask = np.full(self.spectrumK.shape, True, dtype=bool)
                mask[self.picked_wave_index] = False
                self.spectrumK[mask] = 0.0
                self.phase[mask] = 0.0
        else:
            logger.debug("No wave selection has been made, so using all the k nodes")

        # based on the new omega values, calculate the omega spectrum as well
        self.calculate_omega_dispersion()
        if self.spectrum_type == "jonswap":
            self.spectrumW = ms.spectrum_jonswap(
                abs(self.omega_dispersion),
                Hs=self.Hs,
                Tp=self.Tp,
                gamma=self.gamma,
                spectral_version=self.spectral_version,
            )
        elif self.spectrum_type == "gauss":
            self.spectrumW = ms.spectrum_gauss(
                abs(self.omega_dispersion),
                Hs=self.Hs,
                Tp=self.Tp,
                sigma=self.sigma,
                spectral_version=self.spectral_version,
            )
        else:
            raise AssertionError(
                "spectrum type must either be 'jonswap' or 'gauss'. Found {}".format(
                    self.spectrum_type
                )
            )

        (
            self.iW_low,
            self.iW_high,
            self.iW_peak,
            self.W_low,
            self.W_high,
            self.W_peak,
            self.a_peakW,
            self.varianceW,
        ) = ms.specspecs(
            self.omega_dispersion,
            self.spectrumW,
            lowlim=self.E_limit_low,
            higlim=self.E_limit_high,
            mirror=self.mirror,
        )

        # calculate the complex amplitudes for the wave nodes and random phase vector
        self.complex_amplitudes = ms.spectrum_to_complex_amplitudes(
            self.kx_nodes,
            spectral_modulus=self.spectrumK,
            phase=self.phase,
            mirror=self.mirror,
        )

        if not self.wave_construction == "FFT":
            # create the maxtrix exp (j * kx_nodes * x_nodes) where kx_nodes * x_nodes
            # is the matrix following from the vector procuct of the vectos k^T and x
            # only calculate this for DFT
            self.exp_matrix_kx = np.exp(
                1j
                * self.kx_nodes.reshape(self.kx_nodes.size, 1)
                * self.xpoints.reshape((1, self.xpoints.size))
            )

    def calculate_wave_surface(self):
        """
        Calculate the wave surface for current time using either DFT or FFT

        Notes
        -----

        The *wave_construction* attribute determines which algorithm is used to
        calculate the the wave surface. The following options are possible

        * DFTpolar: a DFT (Discrete Fourier Transform) is used to calculate the wave
          surface. There are no requirement to the number and shape of the fourier
          nodes stored in the 'complex_amplitudes'.
        * FFT: a FFT (Fast Fourier Transform) is used to calculate the wave surface.
          This implies that the spectral nodes must be symmetrical around zero as
          S(k)=S^*(-k)
        * DFTcartesian: A DFT is used to calculate the wave surface, however, still the
          same symmetric spectrum as used with the FFT is supplied. This allows to
          compare the DFT and DFT in calculation time and outcome with the exact same
          outcome (as the input nodes can be the same). The scaling with 0.5 due to the
          double spectrum (as it is symmetric) is taken care of here.
        """

        if self.wave_construction in ("DFTpolar", "DFTcartesian"):
            self.amplitude = self.dft_complex_amplitudes(
                self.complex_amplitudes,
                self.exp_matrix_kx,
                self.omega_dispersion,
                self.time,
            )
            if self.wave_construction == "DFTcartesian":
                # The DFTcartesian uses a symmetric spectrum, just as you do with the
                # FFT. Therefore you have to scale the energy with a half
                self.amplitude *= 0.5
        elif self.wave_construction == "FFT":
            # the fft is used
            self.amplitude = self.fft_amplitude(
                self.complex_amplitudes, self.omega_dispersion, self.time
            )
        else:
            raise (
                AssertionError(
                    "wave_construction should be either FFT, DFTpolar, or DFTcartesian."
                    " Found {}".format(self.wave_construction)
                )
            )

        logger.debug(f"H_s of 1D surface {4 * np.std(self.amplitude)} ")

    @staticmethod
    def dft_complex_amplitudes(S_tilde, exp_kx, omega, time):
        """
        Calculate the wave height from the complex amplitudes at given time

        Parameters
        ----------
        S_tilde: complex ndarray
            vector Nx1 with the complex amplitudes
        exp_kx: ndarray array
            NxM matrix with exp(k * x) where k is Nx1 k waves vectors and x is 1XM
            spatial nodes
        omega: ndarray
            N ndvector
        time: float
            Current time
        Returns
        -------
        ndarray array
            Mx1 real array with the wave amplitude for given time

        Notes
        -----
        Calculate the vector height(x0,x1,..xM) = real(sum_k A*exp(j(k*x-w*t))), where
        the MxN matrix exp(k*x) and the Nx1 vector A (with the complex amplitudes) have
        been pre-calculated and exp(j*w*t) is put here in a NxN diagonal matrix.
        Return the Mx1 vector with height for each position x
        """
        N = S_tilde.size
        omega_matrix = np.eye(N) * np.exp(-1j * omega * time)
        M = np.dot(S_tilde, omega_matrix)
        dft = np.dot(M, exp_kx)
        return np.real(dft)

    @staticmethod
    def fft_amplitude(S_tilde, omega, time):
        """Calculate the amplitude at time using the FFT

        Parameters
        ----------
        S_tilde: ndarray
            Nx1 array with the complex amplitudes of the wave spectrum
        omega: ndarray
            N x 1 array with the real angular frequency rad/s per wave vector k
        time:  float
            Current time

        Returns
        -------
        ndarray
            real Nx1 array with the wave amplitude at time t

        Notes
        -----
        Since the FFT is used, the  S_tilde should by symmetrical around k=0 such that
        S(k)=S^*(-k)

        """
        N = S_tilde.size / 2
        ampl = N * np.fft.ifft(S_tilde * np.exp(1j * (-time * omega)))
        # ampl should be real already because S_tilde should by symmetrical around k=0
        # S(k)=S^*(-k) to be sure, take the real value only
        return np.real(ampl)



try:
    import pint
except ImportError as err:
    print("WARNING: {}".format(err))
else:
    ureg = pint.UnitRegistry()

    # define shortcut for quantity specifier using Q_. e.g. Q_("1 m/s") define 1 meter/second
    Q_ = ureg.Quantity



"""
Collection of functions dealing with the system file and directories
"""

import errno
import logging
import os

import pandas as pd

from pymarine.utils.misc import (
    clear_path,
    get_regex_pattern,
    get_time_stamp_from_string,
)

MSG_FORMAT = "{:30s} : {}"

logger = logging.getLogger(__name__)


def get_path_depth(path_name):
    r"""
    Get the depth of a path or file name

    Parameters
    ----------
    path_name : str
        Path name to get the depth from

    Returns
    -------
    int
        depth of the path


    Examples
    --------

    >>> get_path_depth("C:\Anaconda")
    1
    >>> get_path_depth("C:\Anaconda\share")
    2
    >>> get_path_depth("C:\Anaconda\share\pywafo")
    3
    >>> get_path_depth(".\imaginary\path\subdir\share")
    4
    """

    if os.path.isfile(path_name) and os.path.exists(path_name):
        current_path = os.path.split(path_name)[0]
    else:
        current_path = path_name
    depth = 0
    previous_path = current_path
    while current_path not in ("", "."):
        current_path = os.path.split(current_path)[0]
        if current_path == previous_path:
            # For a full path name we end at the root 'C:\'.
            # Detect that by comparing with the previous round
            break
        previous_path = current_path
        depth += 1

    return depth


def scan_base_directory(
    walk_dir=".",
    supplied_file_list=None,
    file_has_string_pattern="",
    file_has_not_string_pattern="",
    dir_has_string_pattern="",
    dir_has_not_string_pattern="",
    start_date_time=None,
    end_date_time=None,
    time_zone=None,
    time_stamp_year_first=True,
    time_stamp_day_first=False,
    extension=None,
    max_depth=None,
    sort_file_base_names=False,
):
    """Recursively scan the directory `walk_dir` and get all files underneath obeying
    the search strings and/or date/time ranges

    Parameters
    ----------
    walk_dir : str, optional
        The base directory to start the import. Default = "."
    supplied_file_list: list, optional
        In case walk dir is not given we can explicitly pass a file list to analyze.
        Default = None
    dir_has_string_pattern : str, optional
        Requires the directory name to have this pattern (Default value = "").
        This selection is only made on the first directory level below the walk_dir
    dir_has_not_string_pattern : str, optional
        Requires the directory name NOT to have this pattern (Default value = "").
        This selection is only made on the first directory level below the walk_dir
    file_has_string_pattern : str, optional
        Requires the file name to have this pattern (Default value = "", i.e.,
        matches all)
    file_has_not_string_pattern : str, optional
        Requires the file name NOT to have this pattern (Default value = "")
    extension : str or None, optional
        Extension of the file to match. If None, also matches. Default = None
    max_depth : int, optional
        Sets a maximum depth to which the search is carried out. Default = None, which
        does not limit the search depth. For deep file structures setting a limit to the
        search depth speeds
        up the search.
    sort_file_base_names: bool, option
        If True, sort the resulting file list alphabetically based on the file base
        name. Default = False
    start_date_time: DateTime or None, optional
        If given, get the date time from the current file name and only add the files
        with a date/time equal or large the *start_date_time*. Default is None
    end_date_time: DateTime or None, optional
        If given, get the date time from the current file name and only add the files
        with a date/time smaller than the *end_date_time*. Default is None
    time_zone:str or None, optional
        If given add this time zone to the file stamp. The start and end time should
        also have a time zone
    time_stamp_year_first: bool, optional
        Passed to the datetime parser. If true, the year is first in the date/time
        string. Default = True
    time_stamp_day_first: bool, optional
        Passed to the datetime parser. If true, the day is first in the date/time
        string. Default = False

    Returns
    -------
    list
        All the file names found below the input directory `walk_dir` obeying all the
        search strings

    Examples
    --------

    Find all the python files under the share directory in the Anaconda installation
    folder

    >>> scan_dir = "C:\\Anaconda\\share"
    >>> file_list = scan_base_directory(scan_dir, extension='.py')

    Find all the python files under the share directory in the Anaconda installation
    folder belonging to the pywafo directory

    >>> file_list = scan_base_directory(scan_dir, extension='.py',
    ...                                 dir_has_string_pattern="wafo")

    Note that wafo matches on the directory 'pywafo', which is the first directory level
    below the scan directory. However, if we would match on '^wafo' the returned list
    would be empty as the directory has to *start* with wafo.

    To get all the files with "test" in the name with a directory depth smaller than
    three do:

    >>> file_list = scan_base_directory(scan_dir, extension='.py',
    ...                                 dir_has_string_pattern="wafo",
    ...                                 file_has_string_pattern="test", max_depth=3)


    Test the date/time boundaries. First create a file list from 28 sep 2017 00:00 to
    5:00 with a hour interval and convert it to a string list

    >>> file_names = ["AMS_{}.mdf".format(dt.strftime("%y%m%dT%H%M%S")) for dt in
    ...    pd.date_range("20170928T000000", "20170928T030000", freq="30min")]
    >>> for file_name in file_names:
    ...     print(file_name)
    AMS_170928T000000.mdf
    AMS_170928T003000.mdf
    AMS_170928T010000.mdf
    AMS_170928T013000.mdf
    AMS_170928T020000.mdf
    AMS_170928T023000.mdf
    AMS_170928T030000.mdf

    Use the scan_base_directory to get the files within a specific date/time range

    >>> file_selection = scan_base_directory(supplied_file_list=file_names,
    ...  start_date_time="20170928T010000", end_date_time="20170928T023000")

    >>> for file_name in file_selection:
    ...     print(file_name)
    AMS_170928T010000.mdf
    AMS_170928T013000.mdf
    AMS_170928T020000.mdf

    Note that the selected range run from 1 am until 2 am; the end_date_time of 2.30 am
    is not included

    """

    # get the regular expression for the has_pattern and has_not_pattern of the files
    # and directories
    file_has_string = get_regex_pattern(file_has_string_pattern)
    file_has_not_string = get_regex_pattern(file_has_not_string_pattern)
    dir_has_string = get_regex_pattern(dir_has_string_pattern)
    dir_has_not_string = get_regex_pattern(dir_has_not_string_pattern)
    logger.debug(MSG_FORMAT.format("file_has_string", file_has_string))
    logger.debug(MSG_FORMAT.format("file_has_not_string", file_has_not_string))
    logger.debug(MSG_FORMAT.format("dir_has_string", dir_has_string))
    logger.debug(MSG_FORMAT.format("dir_has_not_string", dir_has_not_string))

    # use os.walk to recursively walk over all the file and directories
    top_directory = True
    file_list = list()
    logger.debug(f"Scanning directory {walk_dir}")
    for root, subdirs, files in os.walk(walk_dir):
        if supplied_file_list is not None:
            root = "."
            subdirs[:] = list()
            files = supplied_file_list

        logger.debug(f"root={root}  sub={subdirs} files={files}")
        logger.debug(MSG_FORMAT.format("root", root))
        logger.debug(MSG_FORMAT.format("sub dirs", subdirs))
        logger.debug(MSG_FORMAT.format("files", files))
        # get the relative path towards the top directory (walk_dir)
        relative_path = os.path.relpath(root, walk_dir)

        depth = get_path_depth(relative_path)

        if root == walk_dir:
            top_directory = True
        else:
            top_directory = False

        # Base on the first directory list we are going to make a choice of directories
        # to process
        if top_directory:
            include_dirs = list()
            for subdir in subdirs:
                add_dir = False
                if dir_has_string is None or bool(dir_has_string.search(subdir)):
                    add_dir = True
                if add_dir and dir_has_not_string is not None:
                    if bool(dir_has_not_string.search(subdir)):
                        add_dir = False
                if add_dir:
                    include_dirs.append(subdir)
                # Overrule the subdirectory list of os.walk:
                # http://stackoverflow.com/questions/19859840/
                #   excluding-directories-in-os-walk
                logger.debug(f"Overruling subdirs with {include_dirs}")
                subdirs[:] = include_dirs

        for filename in files:
            (filebase, ext) = os.path.splitext(filename)
            if extension is None or extension == ext:
                add_file = False

                if file_has_string is None or bool(file_has_string.search(filebase)):
                    # if has_string is none, the search pattern was either empty or
                    # invalid (which happens during typing the regex in the edit_box).
                    # In this case, always add the file.
                    # If not none, filter on the regex, so only add the file if the
                    # search pattern is in the filename
                    add_file = True

                # Do not add the file in case the has_not string edit has been set
                # (!="") and if the file contains the pattern
                if add_file and file_has_not_string is not None:
                    if bool(file_has_not_string.search(filebase)):
                        # in case we want to exclude the file, the has_not search
                        # pattern must be valid so may not be None
                        add_file = False

                if add_file and (
                    start_date_time is not None or end_date_time is not None
                ):
                    # We have supplied a start time or a end time. See if we can get a
                    # date time from the file name
                    file_time_stamp = get_time_stamp_from_string(
                        string_with_date_time=filebase,
                        yearfirst=time_stamp_year_first,
                        dayfirst=time_stamp_day_first,
                        timezone=time_zone,
                    )

                    if file_time_stamp is not None:
                        # we found a file time stamp. Compare it with the start time
                        if start_date_time is not None:
                            if isinstance(start_date_time, str):
                                # in case the start time was supplied as a string
                                start_date_time = get_time_stamp_from_string(
                                    string_with_date_time=start_date_time,
                                    yearfirst=time_stamp_year_first,
                                    dayfirst=time_stamp_day_first,
                                    timezone=time_zone,
                                )

                            if file_time_stamp < start_date_time:
                                # the file time stamp is smaller, so don't add it
                                add_file = False
                        # if a end time is supplied. Also compare it with the end time
                        if end_date_time is not None:
                            if isinstance(end_date_time, str):
                                end_date_time = get_time_stamp_from_string(
                                    string_with_date_time=end_date_time,
                                    yearfirst=time_stamp_year_first,
                                    dayfirst=time_stamp_day_first,
                                    timezone=time_zone,
                                )
                            if file_time_stamp >= end_date_time:
                                # the file time stamp is larger, so don't add it
                                add_file = False

                if dir_has_string is not None and top_directory:
                    # in case we have specified a directory name with a string search,
                    # exclude the top directory
                    add_file = False

                if max_depth is not None and depth > max_depth:
                    add_file = False

                # create the full base name file
                file_name_to_add = os.path.join(walk_dir, relative_path, filebase)

                # get the path to the stl relative to the selected scan directory
                if add_file:
                    logger.debug(f"Adding file {filebase}")
                    file_list.append(clear_path(file_name_to_add + ext))

    # Sort on the file name. First split the file base from the path, because if the
    # files are in different directories, the first file is not necessarily the oldest
    if sort_file_base_names:
        df = pd.DataFrame(
            data=file_list,
            index=[os.path.split(f)[1] for f in file_list],
            columns=["file_list"],
        )
        df.sort_index(inplace=True)
        file_list = df.file_list.values

    return file_list


def make_directory(directory):
    """Create a directory in case it does not yet exist.

    Parameters
    ----------
    directory : str
        Name of the directory to create

    Notes
    -----
    This function is used to create directories without checking if it already exists.
    If the directory already exists, we can silently continue.

    Raises
    ------
    OSError
        The OSError is only raised if it is not an `EEXIST` error. This implies that the
         creation of the directory failed due to another reason then the directory
         already being present.
         It could be that the file system is full or that we may not have write
         permission

    """
    try:
        os.makedirs(directory)
        logger.debug(f"Created directory : {directory}")
    except OSError as exc:
        # an OSError was raised, see what is the cause
        if exc.errno == errno.EEXIST and os.path.isdir(directory):
            # the output directory already exists, that is ok so just continue
            logger.debug(
                "Directory {} already exists. No problem, we just continue".format(
                    directory
                )
            )
        else:
            # something else was wrong. Raise an error
            logger.warning(
                "Failed to create the directory {} because raised:\n{}".format(
                    directory, exc
                )
            )
            raise



"""
Some miscellaneous functions
"""

import logging
import os
import pathlib
import re
import subprocess
import sys
import time

import dateutil.parser as dparser
import numpy as np
import pandas as pd
import yaml
from numpy.testing import assert_equal

from pymarine.utils import Q_

logger = logging.getLogger(__name__)


class Chdir:
    """Class to move to a directory, do something, and move back when done

    Parameters
    ----------
    new_path: str
        Location where you want to do something

    Notes
    -----
    Used on the Gompute cluster in the batch processing script to submit a job inside a
    directory and then move back to the higher directory to move to the next case

    Examples
    --------

    Go to a known directory (C:/)

    >>> os.chdir("C:/")
    >>> os.getcwd()
    'C:\\\\'

    With the Chdir command we move to the C:/Temp directory where we can do something.

    >>> with Chdir("C:/Temp") as d:
    ...    # in this block we can do something in the directory Temp.
    ...    os.getcwd()
    'C:\\\\Temp'

    We have left the block under Chdir, so we are back at the directory where we started

    >>> os.getcwd()
    'C:\\\\'

    """

    def __init__(self, new_path):
        self.newPath = new_path

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)
        return self

    def __exit__(self, *args):
        os.chdir(self.savedPath)


class Timer:
    """Class to measure the time it takes execute a section of code

    Parameters
    ----------
    message : str
        a string to use to the output line
    name : str, optional
        The name of the routine timed.
    verbose : bool, optional
        if True, produce output
    units : str, optional
        time units to use. Default  'ms'
    n_digits : int, optional
        number of decimals to add to the timer units

    Example
    -------

    Use a `with` / `as` construction to enclose the section of code which need to be
    timed

    >>> from numpy import allclose
    >>> number_of_seconds = 1.0
    >>> with Timer(units="s", n_digits=0) as timer:
    ...    time.sleep(number_of_seconds)
    Elapsed time         routine              :          1 s
    >>> allclose(number_of_seconds, timer.secs, rtol=0.1)
    True
    """

    def __init__(
        self,
        message="Elapsed time",
        name="routine",
        verbose=True,
        units="ms",
        n_digits=0,
        field_width=20,
    ):
        self.verbose = verbose
        self.message = message
        self.name = name
        self.units = units
        self.secs = None

        # Build the format string. E.g. for field_with=20 and n_digits=1 and units=ms,
        # this produces the following
        # "{:<20s} : {:<20s} {:>10.1f} ms"
        self.format_string = (
            "{:<"
            + f"{field_width}"
            + "s}"
            + " {:<"
            + f"{field_width}"
            + "s} : {:>"
            + f"{10}.{n_digits}"
            + "f}"
            + f" {self.units}"
        )

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()

        # start and end are in seconds. Convert time delta to nano seconds
        self.delta_time = np.timedelta64(int(1e9 * (self.end - self.start)), "ns")

        self.secs = float(self.delta_time / np.timedelta64(1, "s"))

        # debug output
        logger.debug(f"Found delta time in ns: {self.delta_time}")

        if self.verbose:
            # convert the delta time to the desired units
            duration = self.delta_time / np.timedelta64(1, self.units)

            # produce output
            logger.info(
                self.format_string.format(self.message, self.name, duration, self.units)
            )


def is_exe(fpath):
    """Test if a file is an executable

    Parameters
    ----------
    fpath : str
        return true or false:

    Returns
    -------
    bool
        In case `fpath` is a file that can be executed return True, else False

    Notes
    -----
    This function can only be used on Linux file systems as the `which` command is used
    to identity the location of the program.
    """
    # use system command 'which' to locate the full location of the file
    p = subprocess.Popen(
        f"which {fpath}",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    p_list = p.stdout.read().splitlines()
    if p_list:
        # which return a path so copy it to fpath
        fpath = p_list[0]
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def clear_path(path_name):
    """routine to clear spurious dots and slashes from a path name
     example bla/././oke becomes bla/oke

    Parameters
    ----------
    path_name :
        return: clear_path as a string

    Returns
    -------
    type
        clear_path as a string

    Examples
    --------

    >>> long_path = os.path.join(".", "..", "ok", "yoo", ".", ".", "") + "/"
    >>> print(long_path)
    .\\..\\ok\\yoo\\.\\.\\/
    >>> print(clear_path(long_path))
    ..\\ok\\yoo

    """
    return str(pathlib.PurePath(path_name))


def delete_module(modname, paranoid=None):
    """Delete a module from memory which loaded before

    Parameters
    ----------
    modname : str
        The name of the module to remove

    paranoid : list or None
         (Default value = None)
    """
    from sys import modules

    try:
        thismod = modules[modname]
    except KeyError:
        raise ValueError(modname)
    these_symbols = dir(thismod)
    if paranoid:
        try:
            paranoid[:]  # sequence support
        except IndexError:
            raise ValueError("must supply a finite list for paranoid")
        else:
            these_symbols = paranoid[:]
    del modules[modname]
    for mod in list(modules.values()):
        try:
            delattr(mod, modname)
        except AttributeError:
            pass
        if paranoid:
            for symbol in these_symbols:
                if symbol[:2] == "__":  # ignore special symbols
                    continue
                try:
                    delattr(mod, symbol)
                except AttributeError:
                    pass


def get_clean_version(version):
    r"""turns the full version string into a clean one without the build

    Parameters
    ----------
    version : str
        The version string as return from versioneer.

    Returns
    -------
    str
        The clean version string

    Notes
    -----
    The version string matches the following regular expression

        "([.|\d]+)([+]*)(.*)"

    This function return the clean version string given by the part "([.|\d]+)"

    Examples
    --------

    >>> get_clean_version("1.3")
    '1.3'
    >>> get_clean_version("2.5+dev.g43429")
    '2.5'
    >>> get_clean_version("4.3.1+dev.g43429-dirty")
    '4.3.1'

    """
    match = re.search(r"([.|\d]+)([+]*)(.*)", version)
    if bool(match):
        version = match.group(1)
    else:
        version = version
    return version


def read_settings_file(file_name):
    """Read the yaml file to get the setup information.

    Parameters
    ----------
    file_name : str
        Name of the configuration file. Can be a full path name as well

    Returns
    -------
    dict
        All the settings as obtained from the yaml configuration file

    Notes
    -----
    The file name of the yaml file is searched for in the following order

    1. The current directory where the script is executed. If a full path is given,
       this will be accepted too.
    2. The directory where the original script is located.

    In this way, a default settings file can be put in the script directory and the user
    does not need to copy it except a setting values needs to be changed

    Raises
    ------
    AssertionError:
        In case the file cannot be found
    """

    if os.path.exists(file_name):
        logger.info(f"Loading configuration file {file_name}")
        configuration_file = file_name
    else:
        logger.info(f"Loading configuration file from script dir {__name__}")
        configuration_file = os.path.join(
            os.path.split(__file__)[0], os.path.split(file_name)[1]
        )
    try:
        logger.debug(f"Trying to read configuration file {configuration_file}")
        with open(configuration_file) as stream:
            settings = yaml.load(stream=stream)
    except OSError as err:
        raise AssertionError(
            "Configuration file can not be found in either current directory of "
            "script directory. Goodbye. {}".format(err)
        )

    return settings


def get_python_version_number(version_info):
    """Script to turn the version info as obtained with sys.version_info into a digit
    number

    Parameters
    ----------
    version_info :
        return: a string with the current python version as a clear digit, i.e., 3.5.3

    Returns
    -------
    type
        a string with the current python version as a clear digit, i.e. 3.5.3

    Examples
    --------
    >>> version_string = get_python_version_number(sys.version_info)

    """

    python_version = f"{version_info.major:d}"
    if version_info.minor != "":
        python_version += f".{version_info.minor:d}"
    if version_info.micro != "":
        python_version += f".{version_info.micro:d}"

    return python_version


def get_regex_pattern(search_pattern):
    """Routine to turn a string into a regular expression which can be used to match a
    string

    Parameters
    ----------
    search_pattern : str
        A regular expression in the form of a string

    Returns
    -------
    None or compiled regular expression
        A regular expression as return by the re.compile function or None in case a
        invalid regular expression was given

    Notes
    -----
    An empty string or an invalid search_pattern will yield a None return

    """
    regular_expression = None
    if search_pattern != "":
        try:
            regular_expression = re.compile(f"{search_pattern}")
        except re.error:
            regular_expression = None
    return regular_expression


def clear_argument_list(argv):
    """
    Small utility to remove the \'\\\\r\' character from the last argument of the argv
    list appearing in cygwin

    Parameters
    ----------
    argv : list
        The argument list stored in `sys.argv`

    Returns
    -------
    list
        Cleared argument list

    """
    new_argv = list()
    for arg in argv:
        # replace the '\r' character with a empty space
        arg = re.sub("\r", "", arg)
        if arg != "":
            # only add the argument if it is not empty
            new_argv.append(arg)
    return new_argv


def query_yes_no(question, default_answer="no"):
    """Ask a yes/no question via raw_input() and return their answer.

    Parameters
    ----------
    question : str
        A question to ask the user
    default_answer : str, optional
        A default answer that is given when only return is hit. Default to 'no'

    Returns
    -------
    str:
        "yes" or "no", depending on the input of the user
    """
    valid = {"yes": "yes", "y": "yes", "ye": "yes", "no": "no", "n": "no"}
    if not default_answer:
        prompt = " [y/n] "
    elif default_answer == "yes":
        prompt = " [Y/n] "
    elif default_answer == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default_answer)

    while 1:
        # sys.stdout.write(question + prompt)
        logger.warning(question + prompt)
        choice = input().lower()
        if default_answer is not None and choice == "":
            return default_answer
        elif choice in list(valid.keys()):
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def print_banner(
    title,
    top_symbol="-",
    bottom_symbol=None,
    side_symbol=None,
    width=80,
    to_stdout=False,
    no_top_and_bottom=False,
):
    """Create a banner for plotting a bigger title above each section in the log output

    Parameters
    ----------
    title :
        The title to plot
    top_symbol : str
        the symbol used for the top line. Default value = "-"
    bottom_symbol : str
        the symbol used for the bottom line. Assume same as top if None is given
        (Default value = None)
    side_symbol : str
        The side symbol. Assume same as top if None is given, except if top is -, then
        take | (Default value = None)
    width : int
        the width of the banner (Default value = 80)
    no_top_and_bottom : bool
        make a simple print without the top and bottom line (Default value = False)
    to_stdout : bool, optional
        Print the banner to the standard output of the console instead of the logging
        system. Defaults to False

    Examples
    --------

    >>> print_banner("This is the start of a section")
    <BLANKLINE>
    --------------------------------------------------------------------------------
    | This is the start of a section                                               |
    --------------------------------------------------------------------------------

    Notes
    -----

    Unless the option 'to_stdout' is set to True, the banner is printed via the logging
    system. Therefore, a logger needs to be created first
    """

    logger.debug("message debug in print_banner")

    if bottom_symbol is None:
        bottom_symbol = top_symbol

    if side_symbol is None:
        if bool(re.match("-", top_symbol)):
            side_symbol = "|"
        else:
            side_symbol = top_symbol

    if not no_top_and_bottom:
        message_string = (
            "{}\n"
            + f"{side_symbol} "
            + "{:"
            + f"{width - 4:d}"
            + "}"
            + f" {side_symbol}"
            + "\n{}"
        )
        message = message_string.format(
            top_symbol * width, title, bottom_symbol * width
        )
    else:
        message_string = f"{side_symbol} " + "{:" + f"{width - 4:d}" + "}"
        message = message_string.format(title)
    if to_stdout:
        print(f"\n{message}")
        sys.stdout.flush()
    else:
        logger.info(f"\n{message}")


def move_script_path_to_back_of_search_path(script_file, append_at_the_end=True):
    """Move the name of a script to the front or the back of the search path

    Parameters
    ----------
    script_file : str
        Name of the script to move

    append_at_the_end: bool, optional, default=True
        Append the name of the script to the end. In case this flag is false, the script
        file is added to the beginning of the path

    Returns
    -------
    list:
        The new system path stored in a list

    Notes
    -----
    This script is sometimes required if the __version string is messing up with
    another __version string

    Examples
    --------

    sys.path = move_script_path_to_back_of_search_path(__file__)
    """
    script_file = os.path.realpath(script_file)
    path_to_script = os.path.split(script_file)[0]
    path_to_script_forward = re.sub("\\\\", "/", path_to_script)
    new_sys_path = list()
    for path in sys.path:
        path_forward = re.sub("\\\\", "/", path)
        if path_forward != path_to_script_forward:
            new_sys_path.append(path)

    if append_at_the_end:
        new_sys_path.append(path_to_script)
    else:
        new_sys_path = [path_to_script] + new_sys_path
    return new_sys_path


def read_value_from_dict_if_valid(dictionary, key, default_value=None):
    """Read a value from a dictionary. If the value is not set, just return the
    default value

    Parameters
    ----------
    dictionary :
        dictionary which is supposed to carry this key field
    key :
        the name of the field to read the value from
    default_value :
        default value in case we fail to read the key field (if it does not exist)

    Returns
    -------
    type
        value of the key field or the default value

    """
    try:
        value = dictionary[key]
    except KeyError:
        value = default_value

    return value


def set_value_if_valid(value, new_value):
    """small routine to set a value on if it is not none.
    Otherwise, the original value is taken

    Parameters
    ----------
    value :
        the original value which you can pre-define with a default value
    new_value :
        the new value. Only set this if it is not none

    Returns
    -------
    type
        net value or the original if new_value was None

    """

    if new_value is not None:
        value = new_value

    return value


def compare_objects(obj1, obj2, counter=0, max_recursion_depth=4):
    """
    Compare if two object are equal

    Parameters
    ----------
    obj1: class
        first object
    obj2: class
        second object
    counter: int
        Current recursion depth. Keeps track of how many time we have recursively called
        this function
    max_recursion_depth: int
        Maximum depth to which we are comparing the objects.


    Notes
    -----
    * This function compares all the attributes of two objects to see if their values
      are the same
    * An attribute field may be another object which we also want to compare with the
      same attribute of the other object. This is done by recursively calling this
      function again.
    * Due to the recursive call mechanism we may end up in a infinite loop. To prevent
      this, a maximum recursion depth can be given.
    * The test function *test_sequence_tool* of the *sequence_tool_utils* module uses
      this function to compare to *SequenceToolSummary* objects

    Raises
    ------
    AssertionError:
        In case on of the object fields is not equal

    """
    for att in dir(obj1):
        if att.startswith("_"):
            continue
        val1 = getattr(obj1, att)
        val2 = getattr(obj2, att)
        try:
            assert_equal(val1, val2)
        except AssertionError:
            if type(val1) is str or type(val1) is list:
                raise
            counter += 1
            if counter < max_recursion_depth:
                compare_objects(val1, val2)
            else:
                continue


def set_default_dimension(
    parse_value, default_dimension=None, force_default_units=False
):
    """
    Add a *pint* dimension to a value


    Parameters
    ----------
    parse_value: ndarray or str or float
        Value with optional a dimension written in the form of a str. Can be an array or
        list of strings as well
    default_dimension: str
        Required default dimension
    force_default_units: bool
        If true the only allowed dimension is the default dimension. Raise an error in
        case this is not the case. Default = False

    Returns
    -------

    :obj:`Quantity`
        Value with the quantity as give by the default

    Raises
    ------
    AssertionError
        In case the dimension of the *parse_value* argument is not not but:

        1. Its dimensionality is not the same as the dimensionality of the
           *default_dimension*
        2. Its units is not the same as the unit of the *default_dimension* and the
           *force_default_units* flag is set to *True*

    Notes
    -----
    * This function is a add-on to the *pint* module, a package to define, operate and
      manipulate physical quantities: https://pypi.python.org/pypi/Pint.
    * This function is used to add a dimension to a value which is parsed from a text
      file.
    * It is checked if the value given in the text file has dimension already, for
      example that it was given as "1.0 m/s".
    * If a dimension was given already: check if the dimensionality (in this case:
      Length/Time)
      is the same as the dimensionality of the *default_dimension* input argument.
    * In case the input value does not have an explicit dimension, the dimension given
      by *default_dimension* is added to the value.
    * This function works on both scalar and list values

    Examples
    --------

    Assume we want to read input values from a text file as plain numbers and we want
    to add a default dimension of *meter* to it in case the value do not have an
    explicit dimension yet. Just do:

    >>> value_without_dimension = 1.0  # value as we read from the text file
    >>> value_with_dimension = set_default_dimension(value_without_dimension, "meter")
    >>> print(value_with_dimension)
    1.0 meter

    The variable *value_with_dimension* is now a pint quantity which carries the
    dimension meter.

    In case the input variable already has a dimension, we should also be able to use
    this *function*. The only requirement is that the dimensionality is the same.
    So this should work:

    >>> value_with_dimension = set_default_dimension("2.5 meter", "meter")
    >>> print(value_with_dimension)
    2.5 meter

    This should work as well

    >>> value_with_dimension = set_default_dimension("5.0 mm", "meter")
    >>> print(value_with_dimension)
    5.0 millimeter

    But this fails as the dimensionality of the input argument is not corresponding with
    the dimensionality of the default dimension

    >>> try:
    ...    value_with_dimension = set_default_dimension("5.0 mm", "second")
    ... except AssertionError:
    ...    print("This fails because the dimensionality is not the same")
    This fails because the dimensionality is not the same

    This function should also work for arrays and list

    >>> values_without_dimension = np.linspace(0, 1, num=5)
    >>> values_with_dimension = set_default_dimension(values_without_dimension,
    ...                                               "meter/second^2")
    >>> print(values_with_dimension)
    [ 0.    0.25  0.5   0.75  1.  ] meter / second ** 2

    Notes
    -----
    * Hz are not converted to rad/s as expected. Therefore, do not try to use this to
      convert Hz -> rad/s
    * If the input argument *parse_val* is None, a None is returned as output as well

    """
    if default_dimension is not None:
        def_unit_val = Q_(1, default_dimension)
    else:
        def_unit_val = None
    dimensionless_unit_val = Q_("1")
    dimensionless = dimensionless_unit_val.dimensionality

    if parse_value is not None:
        # in case no dimensions are given with the parse_value argument, impose them

        if isinstance(parse_value, (list, tuple, np.ndarray)):
            # To properly deal with arrays and lists, first check if we have one
            v = Q_(parse_value[0])
            # If this is allowed, we have and array. Check the value and dimension of
            # the first element
            if (
                v.dimensionality == dimensionless
                and v.units == dimensionless_unit_val.units
            ):
                if not isinstance(parse_value[0], type(dimensionless_unit_val)):
                    # There are no dimensions. Convert the array, add the dimensions
                    # later
                    ret_val = Q_(parse_value)
                else:
                    # We have added the quantity to the parse_value already. Copy it
                    ret_val = parse_value
            else:
                # The elements have a dimension, so convert the array in a bare array
                # without dimensions and copy the dimension type to the default.
                # Then we can just do the conversion below
                parse_value = np.array([Q_(x).magnitude for x in parse_value])
                ret_val = Q_(parse_value)
                if (
                    def_unit_val is not None
                    and v.dimensionality != def_unit_val.dimensionality
                ):
                    raise AssertionError(
                        "The first value of the array given has a dimension with a "
                        "different dimensionality as the default dimension. "
                        "Found {}. Expected {}"
                        "".format(v.dimensionality, def_unit_val.dimensionality)
                    )

                def_unit_val = v.units
        elif not isinstance(parse_value, type(dimensionless_unit_val)):
            # the parse_value is not yet a quantity objects
            ret_val = Q_(parse_value)
        else:
            # The parser value is a quantity already. Copy it
            ret_val = parse_value

        if (
            ret_val.dimensionality == dimensionless
            and ret_val.units == dimensionless_unit_val.units
        ):
            # if no dimension is given, add the default dimension
            ret_val = Q_(np.asarray(parse_value), default_dimension)
            if ret_val.dimensionality != dimensionless:
                logger.debug(
                    "A dimensionless value was and a default dimension was imposed "
                    "{} -> {}.".format(parse_value, ret_val)
                )
        elif def_unit_val is not None:
            # check if the dimensionality is the same as the def_units
            if ret_val.dimensionality != def_unit_val.dimensionality:
                raise AssertionError(
                    "Value given has a dimension with a different dimensionality as "
                    "the default dimension\nFound {}. Expected {}".format(
                        ret_val.dimensionality, def_unit_val.dimensionality
                    )
                )

        # we want to force the units. Check it
        if force_default_units:
            if ret_val.units != def_unit_val.units:
                raise AssertionError(
                    "The dimensions given to the value do not match the default "
                    "units. \n"
                    "Found {}. Expected {}\nPlease fix or set "
                    "*only_default_units_allowed* to False".format(
                        ret_val.units, def_unit_val.units
                    )
                )
    else:
        # in case a none value is given as input just return none as output
        ret_val = None

    return ret_val


def get_value_magnitude(value, convert_to_base_units=True):
    """
    Get the magnitude of value with *Pint* dimension in terms of its base units or
    return a float if *value* does not have a dimension

    Parameters
    ----------
    value: Quantity or float or None
        A value with a Pint dimension or a normal float. In both cases, the value
        without dimension is returned
    convert_to_base_units: bool, optional
        Before turning the value into a magnitude first turn the quantity into its SI
        base units. Default = True

    Returns
    -------
    float or None
        Magnitude of the value in case a Pint Quantity was added to the input or just
        the value itself.
        If *convert_to_base_units* was set to True the value is first converted to its
        SI base units

    Examples
    --------
    Assume we have a value with a pint dimension

    >>> velocity = Q_("2.5 m/s")
    >>> print("Current velocity with dimension is: {}".format(velocity))
    Current velocity with dimension is: 2.5 meter / second

    We can now get the magnitude of *velocity* using this function as

    >>> velocity_mag = get_value_magnitude(velocity)
    >>> print("Velocity without dimension is: {}".format(velocity_mag))
    Velocity without dimension is: 2.5

    In case the input argument of the *get_value_magnitude* is a float and does not
    have a dimension, the value itself is returned

    >>> velocity_mag2 = get_value_magnitude(velocity_mag)
    >>> print("Velocity without dimension is: {}".format(velocity_mag2))
    Velocity without dimension is: 2.5

    In case we have a dimension in none SI units, the value  is by default first
    converted to its SI base units.

    >>> velocity_knots = Q_("1 knot")
    >>> velocity_mag = get_value_magnitude(velocity_knots)
    >>> print("Velocity {} is converted to its magnitude in m/s: {:.2f}"
    ...       "".format(velocity_knots, velocity_mag))
    Velocity 1 knot is converted to its magnitude in m/s: 0.51

    In case that the *convert_to_base_units* flag is False we just get the magnitude in
    the same units as the input argument

    >>> velocity_knots = Q_("2.5 knot")
    >>> velocity_mag = get_value_magnitude(velocity_knots, convert_to_base_units=False)
    >>> print("Velocity {} is converted to its magnitude in knots: {:.2f}"
    ... "".format(velocity_knots, velocity_mag))
    Velocity 2.5 knot is converted to its magnitude in knots: 2.50

    Notes
    -----
    * This function is used inside other functions in which it is not know before hand
      if an input argument is passed with or without a Pint dimension and we only are
      interested in the magnitude of the value. Use this function to get the magnitude
    """

    try:
        if convert_to_base_units:
            value = value.to_base_units()
        value_mag = value.magnitude
    except AttributeError:
        value_mag = value

    return value_mag


def get_time_stamp_from_string(
    string_with_date_time, yearfirst=True, dayfirst=False, timezone=None
):
    """
    Try to get a date/time stamp from a string

    Parameters
    ----------
    string_with_date_time: str
        The string to analyses
    yearfirst: bool, optional
        if true put the year first. See *dateutils.parser*. Default = True
    dayfirst: bool, optional
        if true put the day first. See *dateutils.parser*. Default = False
    timezone: str or None, optional
        if given try to add this time zone:w

    Returns
    -------
    :obj:`DateTime`
        Pandas data time string

    Examples
    --------

    The date time in the file 'AMSBALDER_160929T000000' is  29 sep 2016 and does not
    have a time zone specification. The returned time stamp does also not have a time
    zone

    >>> file_name="AMSBALDER_160929T000000"
    >>> time_stamp =get_time_stamp_from_string(string_with_date_time=file_name)
    >>> print("File name {} has time stamp {}".format(file_name, time_stamp))
    File name AMSBALDER_160929T000000 has time stamp 2016-09-29 00:00:00

    We can also force to add a time zone. The Etc/GMT-2 time zone is UTC + 2 time zone
    which is the central europe summer time (CEST) or the Europe/Amsterdam Summer time.

    >>> time_stamp =get_time_stamp_from_string(string_with_date_time=file_name,
    ...                                        timezone="Etc/GMT-2")
    >>> print("File name {} has time stamp {}".format(file_name, time_stamp))
    File name AMSBALDER_160929T000000 has time stamp 2016-09-29 00:00:00+02:00

    This time we assume the file name already contains a time zone, 2 hours + UTC.
    Since we already have a time zone, the *timezone* option can only convert the date
    time to the specified time zone.

    >>> file_name="AMSBALDER_160929T000000+02"
    >>> time_stamp =get_time_stamp_from_string(string_with_date_time=file_name,
    ...                                        timezone="Etc/GMT-2")
    >>> print("File name {} has time stamp {}".format(file_name, time_stamp))
    File name AMSBALDER_160929T000000+02 has time stamp 2016-09-29 00:00:00+02:00

    In case the time zone given by the *timezone* options differs with the time zone in
    the file name, the time zone is converted

    >>> file_name="AMSBALDER_160929T000000+00"
    >>> time_stamp =get_time_stamp_from_string(string_with_date_time=file_name,
    ...                                        timezone="Etc/GMT-2")
    >>> print("File name {} has time stamp {}".format(file_name, time_stamp))
    File name AMSBALDER_160929T000000+00 has time stamp 2016-09-29 02:00:00+02:00

    """
    try:
        file_time_stamp = dparser.parse(
            string_with_date_time, fuzzy=True, yearfirst=yearfirst, dayfirst=dayfirst
        )
        file_time_stamp = pd.Timestamp(file_time_stamp)
    except ValueError:
        file_time_stamp = None
    else:
        # we have found a time stamp. See if we have to add a time zone
        if timezone is not None:
            try:
                file_time_stamp = file_time_stamp.tz_localize(timezone)
            except TypeError:
                # a time zone was present already. Then try to convert it
                file_time_stamp = file_time_stamp.tz_convert(timezone)

    return file_time_stamp


class PackageInfo:
    """
    A class to analyse the version properties of this package

    Parameters
    ----------
    module_object: :obj:`Module`
        reference to the module for which want to to store the properties

    """

    def __init__(self, module_object):
        self.module_object = module_object

        self.package_version = None
        self.git_sha = None
        self.python_version = None
        self.build_date = None
        self.bundle_dir = None

        if getattr(sys, "frozen", False):
            # we are running in a bundle
            self.get_bundle_version()
        else:
            self.get_source_version()

    def get_bundle_version(self):
        """
        Get the version of the current package from the _version_frozen module which was
        written by the build_executable script.
        """
        try:
            import _version_frozen
        except ImportError:
            print("Could not load _version_frozen. All stay None")
        else:
            print(f"here with bundle {_version_frozen}")
            self.bundle_dir = sys._MEIPASS
            self.package_version = _version_frozen.VERSIONTAG
            self.git_sha = _version_frozen.GIT_SHA
            self.python_version = _version_frozen.PYTHON_VERSION
            self.build_date = _version_frozen.BUILD_DATE

    def get_source_version(self):
        """
        Get the version of the current package via the versioneer approach

        """
        # we are running in a normal Python environment
        self.bundle_dir = os.path.dirname(os.path.abspath(self.module_object.__file__))
        self.package_version = self.module_object.__version__
        self.git_sha = self.module_object.__git_sha_key__
        self.python_version = get_python_version_number(sys.version_info)
        self.build_date = pd.to_datetime("now").strftime("%Y%m%d")


class ConditionalDecorator:
    """
    Add a decorator to a function only if the condition is True

    Parameters
    ----------
    dec: decorator
        The decorator which you want to add when condition is true
    condition: bool
        Only add the decorator if this condition is True
    """

    def __init__(self, dec, condition):
        self.decorator = dec
        self.condition = condition

    def __call__(self, func):
        if not self.condition:
            # Return the function unchanged, not decorated.
            return func
        return self.decorator(func)



import numpy as np
import pandas as pd
from numpy.testing import assert_equal, assert_almost_equal
from pandas.testing import assert_frame_equal

from pymarine.utils.geographic import import_way_points, LocationCheck


def test_import_way_points():
    try:
        data = import_way_points("../data/madeira_ivory.kml", n_distance_points=None)
    except FileNotFoundError:
        data = import_way_points("data/madeira_ivory.kml", n_distance_points=None)

    # create the expected data frame
    data_expected = pd.DataFrame(
        index=np.array(
            [
                0.0,
                51.35344852,
                103.76100595,
                280.37591193,
                540.1554197,
                718.74267821,
                892.77067385,
                1121.19537128,
                1380.98042693,
                1590.32023099,
                1813.18621139,
                2002.15956558,
                2176.81400435,
                2400.01710602,
                2519.33685128,
            ]
        )
    )
    data_expected.index.name = "travel_distance"
    data_expected["latitude"] = np.array(
        [
            32.4,
            31.8,
            31.09051332,
            28.3588147,
            24.03366368,
            21.10140247,
            18.20281473,
            14.38076727,
            10.24088837,
            7.39676749,
            4.39631756,
            2.64584774,
            1.5183185,
            3.43848364,
            4.90171136,
        ]
    )
    data_expected["longitude"] = np.array(
        [
            -17.18,
            -17.9,
            -18.49808275,
            -19.77673541,
            -20.20678434,
            -19.59538317,
            -19.30659972,
            -19.38683736,
            -18.0310994,
            -15.97084304,
            -13.75368275,
            -11.12936662,
            -8.44639843,
            -5.25718259,
            -3.90157281,
        ]
    )
    data_expected["heading"] = np.array(
        [
            225.80012291,
            216.01231006,
            202.54248937,
            185.2219716,
            168.92178621,
            174.55887537,
            181.17301186,
            162.00194422,
            144.06214121,
            143.38056125,
            123.49090439,
            112.61847796,
            59.03325232,
            42.88156671,
            42.88156671,
        ]
    )

    assert_frame_equal(data, data_expected)


def test_import_way_points_interpolated():
    # the kml data without interpolation and select the first five points
    try:
        data_kml = import_way_points(
            "../data/madeira_ivory.kml", n_distance_points=None
        )
    except FileNotFoundError:
        data_kml = import_way_points("data/madeira_ivory.kml", n_distance_points=None)

    data_kml = data_kml.head(5)
    data_kml_expected = pd.DataFrame(
        index=np.array(
            [0.0, 51.3534485188, 103.761005947, 280.375911929, 540.155419698]
        )
    )
    data_kml_expected.index.name = "travel_distance"
    data_kml_expected["latitude"] = np.array(
        [32.4, 31.8, 31.0905133216, 28.3588147039, 24.0336636847]
    )
    data_kml_expected["longitude"] = np.array(
        [-17.18, -17.9, -18.4980827493, -19.7767354088, -20.206784336]
    )
    data_kml_expected["heading"] = np.array(
        [225.800122908, 216.012310055, 202.54248937, 185.221971597, 168.921786213]
    )
    assert_frame_equal(data_kml, data_kml_expected)

    # the kml data with interpolation and select the first five points
    try:
        data_int = import_way_points(
            "../data/madeira_ivory.kml", n_distance_points=2000
        )
    except FileNotFoundError:
        data_int = import_way_points("data/madeira_ivory.kml", n_distance_points=2000)

    # select only the first 5 rows for testing purpose and remove the empty column
    data_int = data_int.head(5)

    data_expected = pd.DataFrame(
        index=np.array([0.0, 2.44157154192, 3.6625019557, 4.88352875138, 6.10465189836])
    )
    data_expected.index.name = "travel_distance"
    data_expected["heading"] = np.array(
        [225.526534405, 225.528796643, 225.533362888, 225.537926615, 225.542487823]
    )
    data_expected["latitude"] = np.array(
        [32.4, 32.3714285714, 32.3571428571, 32.3428571429, 32.3285714286]
    )
    data_expected["longitude"] = np.array(
        [-17.18, -17.2142857143, -17.2314285714, -17.2485714286, -17.2657142857]
    )

    data_int = data_int[data_expected.columns]

    assert_frame_equal(data_int, data_expected)

    # check if the first location of the interpolated dataframe is the same as the pure kml file
    first_location_kml = [data_kml.loc[0, "latitude"], data_kml.loc[0, "longitude"]]
    first_location_dat = [data_int.loc[0, "latitude"], data_int.loc[0, "longitude"]]
    assert_almost_equal([first_location_kml], [first_location_dat])


def test_location_check():
    location_check = LocationCheck(
        latitude=52.37, longitude=4.89, distance_deviation_allowed=10
    )

    out_of_range = location_check.out_of_range(
        current_latitude=52.31, current_longitude=4.78
    )
    assert_equal(out_of_range, True)
    assert_almost_equal(location_check.distance, 10.03906752260248)

    out_of_range = location_check.out_of_range(
        current_latitude=52.36, current_longitude=4.902
    )
    assert_equal(out_of_range, False)
    assert_almost_equal(location_check.distance, 1.3807122835704022)


def main():
    test_import_way_points()
    test_import_way_points_interpolated()
    test_location_check()


if __name__ == "__main__":
    main()
